import matplotlib.pyplot as plt
import numpy as np


def plot_orbit(states, r_earth=6378.137, elevation_azimuth=None, ax=None, *args, **kwargs):
    """
    This function takes a tensor of states, and plots the orbit, together with the Earth.

    Args:
        - states (``torch.tensor``): a set of len(tsince)x2x3 tensor of states,  where the first row represents the spacecraft
                                     position (in km) and the second the spacecraft velocity (in km/s). Reference frame is TEME.
        - r_earth (``float``): Earth radius in km (used for the plot). Default value is 6378.137 km.
        - elevation_azimuth (``tuple``): tuple of two floats, representing the elevation and azimuth angles of the plot. If None, the default values are used.
        - ax (``matplotlib.axes._subplots.Axes3DSubplot``): 3D axis object
        - *args, **kwargs: additional arguments to be passed to the plot function
    
    Returns:
        - ax (``matplotlib.axes._subplots.Axes3DSubplot``): 3D axis object
    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        #if ax is None, we plot the Earth
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x = np.outer(np.cos(u), np.sin(v))
        y = np.outer(np.sin(u), np.sin(v))
        z = np.outer(np.ones(np.size(u)), np.cos(v))

        ax.plot_surface(x*r_earth, y*r_earth, z*r_earth, color='lightblue', alpha=0.3,label='Earth')
        ax.set_xlabel('X [km]')
        ax.set_ylabel('Y [km]')
        ax.set_zlabel('Z [km]')

        # Plot latitude lines (circles)
        for phi in np.arange(0, np.pi, np.pi/6):
            x_lat = np.cos(u) * np.sin(phi)
            y_lat = np.sin(u) * np.sin(phi)
            z_lat = np.full_like(u, np.cos(phi))
            ax.plot(x_lat*r_earth, y_lat*r_earth, z_lat*r_earth, color='black',alpha=0.2)

        # Plot longitude lines
        for theta in np.arange(0, 2 * np.pi, np.pi/6):  
            x_lon = np.cos(theta) * np.sin(v)
            y_lon = np.sin(theta) * np.sin(v)
            z_lon = np.cos(v)
            ax.plot(x_lon*r_earth, y_lon*r_earth, z_lon*r_earth, color='black',alpha=0.2)
    ax.plot(states[:,0,0].numpy(), states[:,0,1].numpy(), states[:,0,2].numpy(),*args, **kwargs)
    ax.legend()
    ax.set_box_aspect([1,1,1])
    if elevation_azimuth is not None:
        ax.view_init(elev=elevation_azimuth[0], azim=elevation_azimuth[1])
    plt.tight_layout()
    return ax

def plot_tles(tles, file_name=None, figsize = (36,18), show=True, axs=None, return_axs=False, log_yscale=False, *args, **kwargs):
    """
    This function takes a list of tles as input and plots the histograms of some of their elements.

    Inputs
    ----------
    - tles (`list`): list of tles, where each element is a `dsgp4.tle.TLE` object.
    - save_fig (`bool`): boolean variable, if True, figure is saved to a file.
    - file_name (`str`): name of the file (including path) where the plot is to be saved.
    - figsize (`tuple`): figure size.

    Outputs
    ----------
    - ax (`numpy.ndarray`): array of AxesSubplot objects
    """
    #I collect all the six variables from the TLEs:
    mean_motion, eccentricity, inclination, argument_of_perigee, raan, b_star, mean_anomaly, mean_motion_first_derivative, mean_motion_second_derivative = [], [], [], [], [], [], [], [], []
    for tle in tles:
        mean_motion.append(tle.mean_motion)
        eccentricity.append(tle.eccentricity)
        inclination.append(tle.inclination)
        argument_of_perigee.append(tle.argument_of_perigee)
        raan.append(tle.raan)
        b_star.append(tle.b_star)
        mean_anomaly.append(tle.mean_anomaly)
        mean_motion_first_derivative.append(tle.mean_motion_first_derivative)
        mean_motion_second_derivative.append(tle.mean_motion_second_derivative)

    plt.rcParams.update({'font.size': 22})
    if axs is None:
        fig, axs = plt.subplots(3, 3, figsize = figsize)

    axs[0,0].hist(mean_motion, bins = max([100,len(tles)]), *args, **kwargs)
    axs[0,0].set_xlabel('Mean Motion [rad/s]')
    x_min, x_max = min(mean_motion), max(mean_motion)
    axs[0,0].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[0,0].set_yscale('log')
    axs[0,0].grid(True)

    axs[0,1].hist(eccentricity, bins = max([100,len(tles)]), *args, **kwargs)
    axs[0,1].set_xlabel('Eccentricity [-]')
    x_min, x_max = min(eccentricity), max(eccentricity)
    axs[0,1].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[0,1].set_yscale('log')
    axs[0,1].grid(True)

    axs[0,2].hist([i*180/np.pi for i in inclination], bins = max([100,len(tles)]), *args, **kwargs)
    axs[0,2].set_xlabel('Inclination [deg]')
    x_min, x_max = min(inclination)*180/np.pi, max(inclination)*180/np.pi
    axs[0,2].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[0,2].set_yscale('log')
    axs[0,2].grid(True)

    axs[1,0].hist([omega*180/np.pi for omega in argument_of_perigee], bins = max([100,len(tles)]), *args, **kwargs)
    axs[1,0].set_xlabel('Argument of Perigee [deg]')
    x_min, x_max = min(argument_of_perigee)*180/np.pi, max(argument_of_perigee)*180/np.pi
    axs[1,0].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[1,0].set_yscale('log')
    axs[1,0].grid(True)

    axs[1,1].hist([RAAN*180/np.pi for RAAN in raan], bins = max([100,len(tles)]), *args, **kwargs)
    axs[1,1].set_xlabel('RAAN [deg]')
    x_min, x_max = min(raan)*180/np.pi, max(raan)*180/np.pi
    axs[1,1].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[1,1].set_yscale('log')
    axs[1,1].grid(True)

    axs[1,2].hist(b_star, bins = max([100,len(tles)]), *args, **kwargs)
    axs[1,2].set_xlabel('Bstar [1/m]')
    x_min, x_max = min(b_star), max(b_star)
    axs[1,2].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[1,2].set_yscale('log')
    axs[1,2].grid(True)

    axs[2,0].hist([M*180/np.pi for M in mean_anomaly], bins = max([100,len(tles)]), *args, **kwargs)
    axs[2,0].set_xlabel('Mean Anomaly [deg]')
    x_min, x_max = min(mean_anomaly)*180/np.pi, max(mean_anomaly)*180/np.pi
    axs[2,0].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[2,0].set_yscale('log')
    axs[2,0].grid(True)

    axs[2,1].hist(mean_motion_first_derivative, bins = max([100,len(tles)]), *args, **kwargs)
    axs[2,1].set_xlabel('Mean Motion 1st Der [rad/s**2]')
    x_min, x_max = min(mean_motion_first_derivative), max(mean_motion_first_derivative)
    axs[2,1].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[2,1].set_yscale('log')
    axs[2,1].grid(True)

    axs[2,2].hist(mean_motion_second_derivative, bins = max([100,len(tles)]), *args, **kwargs)
    axs[2,2].set_xlabel('Mean Motion 2nd Der [rad/s**3]')
    x_min, x_max = min(mean_motion_second_derivative), max(mean_motion_second_derivative)
    axs[2,2].set_xlim(x_min-(x_max-x_min)*0.05, x_max+(x_max-x_min)*0.05)
    if log_yscale:
        axs[2,2].set_yscale('log')
    axs[2,2].grid(True)

    if file_name is not None:
        fig.savefig(fname = file_name)
    if show and not return_axs:
        plt.show()

    if return_axs:
        return axs
    




import numpy
import torch
from .tle import TLE

#@torch.jit.script
def sgp4(satellite, tsince):
    """
    This function represents the SGP4 propagator. Having created the TLE object, and
    initialized the propagator (using `dsgp4.sgp4.sgp4init`), one can use this method
    to propagate the TLE at future times. The method returns the satellite position and velocity
    in km and km/s, respectively, after `tsince` minutes.

    Args:
        - satellite (``dsgp4.tle.TLE``): TLE object
        - tsince (``int`` or ``float`` or ``list`` or ``torch.tensor``): time to propagate, since the TLE epoch, in minutes (can be a vector or a single element)

    Returns:
        - state (``torch.tensor``): a set of len(tsince)x2x3 tensor of states,  where the first row represents the spacecraft
                                    position (in km) and the second the spacecraft velocity (in km/s). Reference frame is TEME.
    """
    #quick check to see if the satellite has been initialized
    if not isinstance(satellite, TLE):
        raise TypeError('The satellite object should be a dsgp4.tle.TLE object.')
    if not hasattr(satellite, '_radiusearthkm'):
        raise AttributeError('It looks like the satellite has not been initialized. Please use the `initialize_tle` method or directly `sgp4init` to initialize the satellite. Otherwise, if you are propagating, another option is to use `dsgp4.propagate` and pass `initialized=True` in the arguments.')
    #in case an int, float or list are passed, convert them to torch.tensor
    if isinstance(tsince, (int, float, list)):
        tsince = torch.tensor(tsince)
    mrt = torch.zeros(tsince.size())
    temp4 = torch.ones(tsince.size())*1.5e-12
    x2o3  = torch.tensor(2.0 / 3.0)

    vkmpersec    = torch.ones(tsince.size())*(satellite._radiusearthkm * satellite._xke/60.0)

    # sgp4 error flag
    satellite._t    = tsince.clone()
    satellite._error = torch.tensor(0)

    #  secular gravity and atmospheric drag
    xmdf    = satellite._mo + satellite._mdot * satellite._t
    argpdf  = satellite._argpo + satellite._argpdot * satellite._t
    nodedf  = satellite._nodeo + satellite._nodedot * satellite._t
    argpm   = argpdf
    mm     = xmdf
    t2     = satellite._t * satellite._t
    nodem   = nodedf + satellite._nodecf * t2
    tempa   = 1.0 - satellite._cc1 * satellite._t
    tempe   = satellite._bstar * satellite._cc4 * satellite._t
    templ   = satellite._t2cof * t2

    if satellite._isimp != 1:
        delomg = satellite._omgcof * satellite._t
        delmtemp =  1.0 + satellite._eta * xmdf.cos()
        delm   = satellite._xmcof * \
                  (delmtemp * delmtemp * delmtemp -
                  satellite._delmo)
        temp   = delomg + delm
        mm     = xmdf + temp
        argpm  = argpdf - temp
        t3     = t2 * satellite._t
        t4     = t3 * satellite._t
        tempa  = tempa - satellite._d2 * t2 - satellite._d3 * t3 - \
                          satellite._d4 * t4
        tempe  = tempe + satellite._bstar * satellite._cc5 * (mm.sin() -
                          satellite._sinmao)
        templ  = templ + satellite._t3cof * t3 + t4 * (satellite._t4cof +
                          satellite._t * satellite._t5cof)

    nm    = satellite._no_unkozai.clone()
    em    = satellite._ecco.clone()
    inclm = satellite._inclo.clone()

    satellite._error=torch.any(nm<=0)*2

    am = torch.pow((satellite._xke / nm),x2o3) * tempa * tempa
    nm = satellite._xke / torch.pow(am, 1.5)
    em = em - tempe

    if satellite._error==0.:
        satellite._error=torch.any((em>=1.0) | (em<-0.001))*1

    #  sgp4fix fix tolerance to avoid a divide by zero
    em=torch.where(em<1.0e-6,1.0e-6,em)
    mm     = mm + satellite._no_unkozai * templ
    xlm    = mm + argpm + nodem
    emsq   = em * em
    temp   = 1.0 - emsq

    nodem = torch.fmod(nodem, torch.tensor(2*numpy.pi))

    argpm  = argpm % (2*numpy.pi)
    xlm    = xlm % (2*numpy.pi)
    mm     = (xlm - argpm - nodem) % (2*numpy.pi)

    satellite._am = am.clone()
    satellite._em = em.clone()
    satellite._im = inclm.clone()
    satellite._Om = nodem.clone()
    satellite._om = argpm.clone()
    satellite._mm = mm.clone()
    satellite._nm = nm.clone()

    # compute extra mean quantities
    sinim = inclm.sin()
    cosim = inclm.cos()

    # add lunar-solar periodics
    ep     = em
    xincp  = inclm
    argpp  = argpm
    nodep  = nodem
    mp     = mm
    sinip  = sinim
    cosip  = cosim

    axnl = ep * argpp.cos()
    temp = 1.0 / (am * (1.0 - ep * ep))
    aynl = ep* argpp.sin() + temp * satellite._aycof
    xl   = mp + argpp + nodep + temp * satellite._xlcof * axnl

    # solve kepler's equation
    u    = (xl - nodep) % (2*numpy.pi)
    eo1  = u
    tem5 = torch.ones(tsince.size())
    # kepler iteration
    for _ in range(10):
        coseo1=eo1.cos()
        sineo1=eo1.sin()
        tem5   = 1.0 - coseo1 * axnl - sineo1 * aynl
        tem5   = (u - aynl * coseo1 + axnl * sineo1 - eo1) / tem5
        tem5=torch.where(tem5>=0.95, 0.95, tem5)
        tem5=torch.where(tem5<=-0.95, -0.95, tem5)
        eo1    = eo1 + tem5

    #  short period preliminary quantities
    ecose = axnl*coseo1 + aynl*sineo1
    esine = axnl*sineo1 - aynl*coseo1
    el2   = axnl*axnl + aynl*aynl
    pl    = am*(1.0-el2)
    if satellite._error==0.:
        satellite._error=torch.any(pl<0.)*4

    rl     = am * (1.0 - ecose)
    rdotl  = am.sqrt() * esine/rl
    rvdotl = pl.sqrt() / rl
    betal  = (1.0 - el2).sqrt()
    temp   = esine / (1.0 + betal)
    sinu   = am / rl * (sineo1 - aynl - axnl * temp)
    cosu   = am / rl * (coseo1 - axnl + aynl * temp)
    su     = torch.atan2(sinu, cosu)
    sin2u  = (cosu + cosu) * sinu
    cos2u  = 1.0 - 2.0 * sinu * sinu
    temp   = 1.0 / pl
    temp1  = 0.5 * satellite._j2 * temp
    temp2  = temp1 * temp

    mrt   = rl * (1.0 - 1.5 * temp2 * betal * satellite._con41) + \
             0.5 * temp1 * satellite._x1mth2 * cos2u
    su    = su - 0.25 * temp2 * satellite._x7thm1 * sin2u
    xnode = nodep + 1.5 * temp2 * cosip * sin2u
    xinc  = xincp + 1.5 * temp2 * cosip * sinip * cos2u
    mvt   = rdotl - nm * temp1 * satellite._x1mth2 * sin2u / satellite._xke
    rvdot = rvdotl + nm * temp1 * (satellite._x1mth2 * cos2u +
             1.5 * satellite._con41) / satellite._xke

    # orientation vectors
    sinsu =  su.sin()
    cossu =  su.cos()
    snod  =  xnode.sin()
    cnod  =  xnode.cos()
    sini  =  xinc.sin()
    cosi  =  xinc.cos()
    xmx   = -snod * cosi
    xmy   =  cnod * cosi
    ux    =  xmx * sinsu + cnod * cossu
    uy    =  xmy * sinsu + snod * cossu
    uz    =  sini * sinsu
    vx    =  xmx * cossu - cnod * sinsu
    vy    =  xmy * cossu - snod * sinsu
    vz    =  sini * cossu

    # position and velocity (in km and km/sec)
    _mr = mrt * satellite._radiusearthkm

    r = torch.stack((_mr * ux, _mr * uy, _mr * uz))
    v = torch.stack(((mvt * ux + rvdot * vx) * vkmpersec,
          (mvt * uy + rvdot * vy) * vkmpersec,
          (mvt * uz + rvdot * vz) * vkmpersec))

    # decaying satellites
    if satellite._error==0.:
        satellite._error=torch.any(mrt<1.0)*6
    return torch.transpose(torch.stack((r.squeeze(),v.squeeze()),1),0,-1)#torch.cat((r.swapaxes(0,2),v.swapaxes(0,2)),1)#torch.stack(list(r)+list(v)).reshape(2,3)



import numpy as np
import torch
from .sgp4 import sgp4
from .sgp4init import sgp4init
from . import util
from .tle import TLE

def _propagate(x, tle_sat, tsince, gravity_constant_name="wgs-84"):
    from .sgp4init import sgp4init
    from .sgp4 import sgp4
    whichconst=util.get_gravity_constants(gravity_constant_name)
    sgp4init(whichconst=whichconst,
                        opsmode='i',
                        satn=tle_sat.satellite_catalog_number,
                        epoch=(tle_sat._jdsatepoch+tle_sat._jdsatepochF)-2433281.5,
                        xbstar=x[0],
                        xndot=x[1],
                        xnddot=x[2],
                        xecco=x[3],
                        xargpo=x[4],
                        xinclo=x[5],
                        xmo=x[6],
                        xno_kozai=x[7],
                        xnodeo=x[8],
                        satellite=tle_sat)
    state=sgp4(tle_sat, tsince*torch.ones(1,1))
    return state

def initial_guess(tle_0, time_mjd, target_state=None):
    """
    This method takes an initial TLE and the time at which we want to propagate it, and returns
    a set of parameter related to an initial guess useful to find the TLE observation that corresponds to the propagated state

    Args:
        - tle_0 (``dsgp4.tle.TLE``): starting TLE at time0
        - new_date (``datetime.datetime``): new date of the TLE, as a datetime object
        - target_state (``torch.tensor``): a 2x3 tensor for the position and velocity of the state. If None, then this is computed
                                           by propagating `tle_0` at the `new_date`.

    Returns:
        - y0 (``torch.tensor``): initial guess for the TLE elements. In particular, y0 contains
                                 the following elements (see SGP4 for a thorough description of these parameters):
                                * y0[0]: bstar (B* parameter)
                                * y0[1]: ndot (mean motion first derivative)
                                * y0[2]: nddot (mean motion second derivative)
                                * y0[3]: ecco (eccentricity)
                                * y0[4]: argpo (argument of perigee)
                                * y0[5]: inclo (inclination)
                                * y0[6]: mo (mean anomaly)
                                * y0[7]: no_kozai (mean anomaly in certain units)
                                * y0[8]: nodeo (right ascension of the ascending node)
        - target_state (``torch.tensor``): this expresses the cartesian state as position & velocity in km & km/s, at the propagated time.
                                            The objective of the Newton method is to find the TLE observation that corresponds to that, at the propagated
                                            time.
        - new_tle (``dsgp4.tle.TLE``): TLE constructed with `y0`, in order to find the TLE that corresponds to `target_state`
        - tle_elements_0 (``dict``): dictionary used to construct `new_tle`

    """
    new_date=util.from_mjd_to_datetime(time_mjd)
    if target_state is None:
        whichconst=util.get_gravity_constants("wgs-84")
        sgp4init(whichconst=whichconst,
                            opsmode=tle_0._opsmode,
                            satn=tle_0.satellite_catalog_number,
                            epoch=(tle_0._jdsatepoch+tle_0._jdsatepochF)-2433281.5,
                            xbstar=tle_0._bstar,
                            xndot=tle_0._ndot,
                            xnddot=tle_0._nddot,
                            xecco=tle_0._ecco,
                            xargpo=tle_0._argpo,
                            xinclo=tle_0._inclo,
                            xmo=tle_0._mo,
                            xno_kozai=tle_0._no_kozai,
                            xnodeo=tle_0._nodeo,
                            satellite=tle_0)
        tsince=(time_mjd-util.from_datetime_to_mjd(tle_0._epoch))*1440.
        x=tsince*torch.ones(1,1,requires_grad=True)#torch.rand(1,1, requires_grad=True)
        target_state=sgp4(tle_0, x)
    #print(target_state.size(), target_state)
    tle_elements_0=util.from_cartesian_to_tle_elements(target_state.detach().numpy()*1e3)
    #you need to transform it to the appropriate units
    xpdotp=1440.0 / (2.0 *np.pi)
    tle_elements_0['epoch_year']=new_date.year
    tle_elements_0['epoch_days']=util.from_datetime_to_fractional_day(new_date)
    #tle_elements_0['no_kozai']=tle_elements_0['mean_motion']/(np.pi/43200.0)/xpdotp
    #tle_elements_0['inclo']=tle_elements_0['inclination']
    #tle_elements_0['nodeo']=tle_elements_0['raan']
    #tle_elements_0['argpo']=tle_elements_0['argument_of_perigee']
    #tle_elements_0['mo']=tle_elements_0['mean_anomaly']
    #tle_elements_0['ecco']=tle_elements_0['eccentricity']
    tle_elements_0['mean_motion_first_derivative']=tle_0.mean_motion_first_derivative
    tle_elements_0['mean_motion_second_derivative']=tle_0.mean_motion_second_derivative
    tle_elements_0['classification']=tle_0.classification
    tle_elements_0['international_designator']=tle_0.international_designator
    tle_elements_0['satellite_catalog_number']=tle_0.satellite_catalog_number
    tle_elements_0['ephemeris_type']=tle_0.ephemeris_type
    tle_elements_0['element_number']=tle_0.element_number
    tle_elements_0['b_star']=tle_0.b_star
    tle_elements_0['revolution_number_at_epoch']=tle_0.revolution_number_at_epoch
    new_tle=TLE(tle_elements_0)
    y0=torch.tensor([new_tle._bstar,
                     new_tle._ndot,
                     new_tle._nddot,
                     new_tle._ecco,
                     new_tle._argpo,
                     new_tle._inclo,
                     new_tle._mo,
                     new_tle._no_kozai,
                     new_tle._nodeo,
                     0.],requires_grad=True)
    return y0, target_state, new_tle, tle_elements_0

def update_TLE(old_tle,y0):
    """
    This method takes a TLE and an initial guess, and returns a TLE updated accordingly. It
    is useful while doing Newton iterations.

    Args:
        - old_tle (``dsgp4.tle.TLE``): TLE corresponding to previous guess
        - y0 (``torch.tensor``): initial guess (see the docstrings for `initial_guess` to know the content of `y0`)

    Returns:
        - new_tle (``dsgp4.tle.TLE``): updated TLE

    """
    tle_elements={}
    xpdotp=1440.0/(2.0 *np.pi)
    #I need to convert no_kozai to consistent units for what is expected by TLE class:
    no_kozai=float(y0[7])*xpdotp
    tle_elements['mean_motion']=no_kozai*np.pi/43200.0

    #the rest is consistent:
    tle_elements['b_star']=float(y0[0])
    tle_elements['raan']=float(y0[8])
    tle_elements['eccentricity']=float(y0[3])
    tle_elements['argument_of_perigee']=float(y0[4])
    tle_elements['inclination']=float(y0[5])
    tle_elements['mean_anomaly']=float(y0[6])

    #now the ones that stay the same:
    tle_elements['mean_motion_first_derivative']=old_tle.mean_motion_first_derivative
    tle_elements['mean_motion_second_derivative']=old_tle.mean_motion_second_derivative
    tle_elements['epoch_days']=old_tle.epoch_days
    tle_elements['epoch_year']=old_tle.epoch_year
    tle_elements['classification']=old_tle.classification
    tle_elements['satellite_catalog_number']=old_tle.satellite_catalog_number
    tle_elements['ephemeris_type']=old_tle.ephemeris_type
    tle_elements['international_designator']=old_tle.international_designator
    tle_elements['revolution_number_at_epoch']=old_tle.revolution_number_at_epoch
    tle_elements['element_number']=old_tle.element_number
    return TLE(tle_elements)

def newton_method(tle_0, time_mjd, target_state=None, new_tol=1e-12,max_iter=50):
    """
    This method performs Newton method starting from an initial TLE and a given propagation time. The objective
    is to find a TLE that accurately reconstructs the propagated state, at observation time.

    Args:
        - tle_0 (``dsgp4.tle.TLE``): starting TLE (i.e., TLE at a given initial time)
        - new_date (``datetime.datetime``): time (as a datetime object) at which we want the state to be propagated, and we want the TLE at that time
        - target_state (``torch.tensor``): 2x3 tensor representing target state. If None, then this is directly computed by propagating the TLE at `new_date`
        - new_tol (``float``): newton tolerance
        - max_iter (``int``): maximum iterations for Newton's method

    Returns:
        - tle (``dsgp4.tle.TLE``): found TLE
        - y (``torch.tensor``): returns the solution that satisfied F(y)=0 (with a given tolerance)
                                or (in case no convergence is reached within the tolerance) the best
                                guess found in `max_iter` iterations
    """
    i=0
    tol=1e9
    y0, state_target, next_tle, tle_elements_0=initial_guess(tle_0=tle_0,time_mjd=time_mjd,target_state=target_state)
    #print(y0,states)
    #newton iterations:
    while i<max_iter and tol>new_tol:
        #print(f"y0: {y0}")
        propagate=lambda x: _propagate(x,next_tle,(time_mjd-util.from_datetime_to_mjd(next_tle._epoch))*1440.)
        y1=util.clone_w_grad(y0)
        y2=util.clone_w_grad(y0)
        y3=util.clone_w_grad(y0)
        y4=util.clone_w_grad(y0)
        y5=util.clone_w_grad(y0)
        r_x=propagate(y0)[0][0]
        r_x.backward()
        gradient_rx=y0.grad
        r_y=propagate(y1)[0][1]
        r_y.backward()
        gradient_ry=y1.grad
        r_z=propagate(y2)[0][2]
        r_z.backward()
        gradient_rz=y2.grad
        v_x=propagate(y3)[1][0]
        v_x.backward()
        gradient_vx=y3.grad
        v_y=propagate(y4)[1][1]
        v_y.backward()
        gradient_vy=y4.grad
        v_z=propagate(y5)[1][2]
        v_z.backward()
        gradient_vz=y5.grad
        F=np.array([float(r_x)-float(state_target[0][0]),float(r_y)-float(state_target[0][1]), float(r_z)-float(state_target[0][2]), float(v_x)-float(state_target[1][0]), float(v_y)-float(state_target[1][1]), float(v_z)-float(state_target[1][2])])
        tol=np.linalg.norm(F)
        #print(f"|F|: {tol}")
        DF=np.stack((gradient_rx, gradient_ry, gradient_rz, gradient_vx, gradient_vy, gradient_vz))
        #we remove the first three columns (they are all zeros):
        DF=DF[:,3:]
        DF=DF[:,:-1]
        dY=-np.matmul(np.matmul(np.linalg.pinv(np.matmul(DF.T,DF)),DF.T),F)
        #we make sure the eccentricity does not go to negative values:
        if float(y0[3])+dY[0]<0:
            dY[0]=-float(y0[3])*0.9999
        dY=torch.tensor([0.,0.,0.]+list(dY)+[0.], requires_grad=True)
        if tol<new_tol:
            print(f"F(y): {np.linalg.norm(F)}")
            print(f"Solution found, at iter: {i}")
            return next_tle, y0#+dY
        else:
            #Newton update:
            #y0=y0+dY
            y0=torch.tensor([float(el1)+float(el2) for el1, el2 in zip(list(y0),list(dY))],requires_grad=True)
            next_tle=update_TLE(next_tle, y0)
        i+=1
    print("Solution not found, returning best found so far")
    print(f"F(y): {np.linalg.norm(F)}")
    return next_tle, y0



import numpy as np
import torch
from .sgp4_batched import sgp4_batched
from .util import gstime

def initl_batch(
       xke, j2,
       ecco,   epoch,  inclo,   no,
       method,
       opsmode, batch_size
       ):

     x2o3   = torch.full((batch_size,),2.0 / 3.0);

     eccsq  = ecco * ecco;
     omeosq = 1.0 - eccsq;
     rteosq = omeosq.sqrt();
     cosio  = inclo.cos();
     cosio2 = cosio * cosio;

     ak    = torch.pow(xke / no, x2o3);
     d1    = 0.75 * j2 * (3.0 * cosio2 - 1.0) / (rteosq * omeosq);
     del_  = d1 / (ak * ak);
     adel  = ak * (1.0 - del_ * del_ - del_ *
             (1.0 / 3.0 + 134.0 * del_ * del_ / 81.0));
     del_  = d1/(adel * adel);
     no    = no / (1.0 + del_);

     ao    = torch.pow(xke / no, x2o3);
     sinio = inclo.sin();
     po    = ao * omeosq;
     con42 = 1.0 - 5.0 * cosio2;
     con41 = -con42-cosio2-cosio2;
     ainv  = 1.0 / ao;
     posq  = po * po;
     rp    = ao * (1.0 - ecco);
     method = ['n']*batch_size;

     if opsmode == 'a':
         #  gst time
         ts70  = epoch - 7305.0;
         ds70 = torch.floor_divide(ts70 + 1.0e-8,1);
         tfrac = ts70 - ds70;
         #  find greenwich location at epoch
         c1    = torch.tensor(1.72027916940703639e-2);
         thgr70= torch.tensor(1.7321343856509374);
         fk5r  = torch.tensor(5.07551419432269442e-15);
         c1p2p = c1 + (2*numpy.pi);
         gsto  = (thgr70 + c1*ds70 + c1p2p*tfrac + ts70*ts70*fk5r) % (2*numpy.pi)
         if gsto < 0.0:
             gsto = gsto + (2*numpy.pi);

     else:
        gsto = gstime(epoch + 2433281.5);

     return (
       no,
       method,
       ainv,  ao,    con41,  con42, cosio,
       cosio2,eccsq, omeosq, posq,
       rp,    rteosq,sinio , gsto,
       )


def sgp4init_batch(
       whichconst,   opsmode,  satn, epoch,
       xbstar,   xndot,   xnddot,   xecco,   xargpo,
       xinclo,   xmo,   xno_kozai,
       xnodeo,  satellite_batch
       ):
     """
     This function initializes the sgp4 propagator.
     Args:
        - whichconst (``tuple``): this contains all the necessary constants (tumin, mu, radiusearthkm, xke,
                        j2, j3, j4, j3oj2)), according to the chosen gravitational model (wgs-72, wgs-84,
                        wgs-72old are three possible choices)
        - opsmode (``str``): mode of operation (possibilities: afspc or improved, 'a' and 'i', respectively)
        - satn (``str``): satellite catalog number
        - epoch (``torch.float``): TLE days since 1949 December 31 00:00 UT
        - xbstar (``torch.float``): drag coefficient (in 1/earth_radii)
        - xndot (``torch.float``): first derivative of mean motion (in revs/day**2)
        - xnddot (``torch.float``): second derivative mean motion (in revs/day**3)
        - xecco (``torch.float``): eccentricity
        - xargpo (``torch.float``): argument of perigee [rad]
        - xinclo (``torch.float``): inclination [rad]
        - xmo (``torch.float``): mean anomaly [rad]
        - xno_kozai (``torch.float``): mean motion (radians/minute)
        - xnodeo (``torch.float``): right ascension of the ascending node [rad]
        - satellite_batch (``dsgp4.tle.TLE``): TLE object, that represents the batch of TLEs (each element is an N-dimensional tensor)

    ..Note:
        If satellite_batch._error is different than 0 once this routine is called, then the satellite propagation
        has had an error. In particular, the following error values hold for `satellite_batch._error`:
        * `1` -> eccentricity `>=1.` or `<-0.001` or semi-major axis (in Earth radii) `<0.95`
        * `2` -> mean_motion `<0.`
        * `3` -> eccentricity `<0.` or `>1.`
        * `4` -> semi-latus rectum `<0.`
        * `5` -> epoch elements are sub-orbital
        * `6` -> satellite has decayed
     """
     batch_size = len(xecco)
     temp4    =   torch.tensor(1.5e-12)

    #  ----------- set all near earth variables to zero ------------
     satellite_batch._isimp   = torch.full((batch_size,),0);   satellite_batch._method = ["n"]*batch_size;              satellite_batch._aycof    = torch.full((batch_size,),0.0);
     satellite_batch._con41   = torch.full((batch_size,),0.0); satellite_batch._cc1    = torch.full((batch_size,),0.0); satellite_batch._cc4      = torch.full((batch_size,),0.0);
     satellite_batch._cc5     = torch.full((batch_size,),0.0); satellite_batch._d2     = torch.full((batch_size,),0.0); satellite_batch._d3       = torch.full((batch_size,),0.0);
     satellite_batch._d4      = torch.full((batch_size,),0.0); satellite_batch._delmo  = torch.full((batch_size,),0.0); satellite_batch._eta      = torch.full((batch_size,),0.0);
     satellite_batch._argpdot = torch.full((batch_size,),0.0); satellite_batch._omgcof = torch.full((batch_size,),0.0); satellite_batch._sinmao   = torch.full((batch_size,),0.0);
     satellite_batch._t       = torch.full((batch_size,),0.0); satellite_batch._t2cof  = torch.full((batch_size,),0.0); satellite_batch._t3cof    = torch.full((batch_size,),0.0);
     satellite_batch._t4cof   = torch.full((batch_size,),0.0); satellite_batch._t5cof  = torch.full((batch_size,),0.0); satellite_batch._x1mth2   = torch.full((batch_size,),0.0);
     satellite_batch._x7thm1  = torch.full((batch_size,),0.0); satellite_batch._mdot   = torch.full((batch_size,),0.0); satellite_batch._nodedot  = torch.full((batch_size,),0.0);
     satellite_batch._xlcof   = torch.full((batch_size,),0.0); satellite_batch._xmcof  = torch.full((batch_size,),0.0); satellite_batch._nodecf   = torch.full((batch_size,),0.0);

     #  ------------------------ earth constants -----------------------
     #  sgp4fix identify constants and allow alternate values
     #  this is now the only call for the constants
     (satellite_batch._tumin, satellite_batch._mu, satellite_batch._radiusearthkm, satellite_batch._xke,
       satellite_batch._j2, satellite_batch._j3, satellite_batch._j4, satellite_batch._j3oj2) = whichconst

 	 # -------------------------------------------------------------------------

     satellite_batch._error = torch.full((batch_size,),0)
     satellite_batch._operationmode = opsmode
     satellite_batch._satnum = satn

     satellite_batch._bstar   = xbstar.clone()
    # sgp4fix allow additional parameters in the struct
     satellite_batch._ndot    = xndot.clone()
     satellite_batch._nddot   = xnddot.clone()
     satellite_batch._ecco    = xecco.clone()
     satellite_batch._argpo   = xargpo.clone()
     satellite_batch._inclo   = xinclo.clone()
     satellite_batch._mo	    = xmo.clone()
	# sgp4fix rename variables to clarify which mean motion is intended
     satellite_batch._no_kozai= xno_kozai.clone()
     satellite_batch._nodeo   = xnodeo.clone()

    # single averaged mean elements
     satellite_batch._am = torch.full((batch_size,),0.0)
     satellite_batch._em = torch.full((batch_size,),0.0)
     satellite_batch._im = torch.full((batch_size,),0.0)
     satellite_batch._Om = torch.full((batch_size,),0.0)
     satellite_batch._mm = torch.full((batch_size,),0.0)
     satellite_batch._nm = torch.full((batch_size,),0.0)

     ss     = 78.0 / satellite_batch._radiusearthkm + 1.0

     qzms2ttemp = (120.0 - 78.0) / satellite_batch._radiusearthkm
     qzms2t = qzms2ttemp * qzms2ttemp * qzms2ttemp * qzms2ttemp
     x2o3   =  2.0 / 3.0

     satellite_batch._init = 'y'
     satellite_batch._t	 = 0.0

    # sgp4fix remove satn as it is not needed in initl
     (
       satellite_batch._no_unkozai,
       method,
       ainv,  ao,    satellite_batch._con41,  con42, cosio,
       cosio2,eccsq, omeosq, posq,
       rp,    rteosq,sinio , satellite_batch._gsto,
       ) = initl_batch(
           satellite_batch._xke, satellite_batch._j2, satellite_batch._ecco, epoch, satellite_batch._inclo, satellite_batch._no_kozai, satellite_batch._method,
           satellite_batch._operationmode,batch_size
         )
     satellite_batch._a    = torch.pow( satellite_batch._no_unkozai*satellite_batch._tumin , (-2.0/3.0) )
     satellite_batch._alta = satellite_batch._a*(1.0 + satellite_batch._ecco) - 1.0
     satellite_batch._altp = satellite_batch._a*(1.0 - satellite_batch._ecco) - 1.0

     # Assuming all relevant variables are tensors.
     condition1 = (omeosq >= 0.0) | (satellite_batch._no_unkozai >= 0.0)

     # Initialize all variables that will be conditionally assigned
     satellite_batch._isimp = torch.where(condition1, torch.tensor(0), satellite_batch._isimp)
     condition2 = condition1 & (rp < 220.0 / satellite_batch._radiusearthkm + 1.0)
     satellite_batch._isimp = torch.where(condition2, torch.tensor(1), satellite_batch._isimp)

     sfour = torch.where(condition1, ss, torch.zeros_like(ss))
     qzms24 = torch.where(condition1, qzms2t, torch.zeros_like(qzms2t))
     perige = torch.where(condition1, (rp - 1.0) * satellite_batch._radiusearthkm, torch.zeros_like(rp))

     condition3 = condition1 & (perige < 156.0)
     sfour_temp1 = torch.where(perige < 98.0, torch.tensor(20.0), perige - 78.0)
     sfour = torch.where(condition3, sfour_temp1, sfour)
     qzms24temp = (120.0 - sfour) / satellite_batch._radiusearthkm
     qzms24 = torch.where(condition3, qzms24temp ** 4, qzms24)
     sfour = torch.where(condition3, sfour / satellite_batch._radiusearthkm + 1.0, sfour)

     pinvsq = 1.0 / posq
     tsi = 1.0 / (ao - sfour)
     satellite_batch._eta = ao * satellite_batch._ecco * tsi
     etasq = satellite_batch._eta ** 2
     eeta = satellite_batch._ecco * satellite_batch._eta
     psisq = torch.abs(1.0 - etasq)
     coef = qzms24 * (tsi ** 4)
     coef1 = coef / (psisq ** 3.5)
     cc2 = coef1 * satellite_batch._no_unkozai * (ao * (1.0 + 1.5 * etasq + eeta *
                    (4.0 + etasq)) + 0.375 * satellite_batch._j2 * tsi / psisq * satellite_batch._con41 *
                    (8.0 + 3.0 * etasq * (8.0 + etasq)))
     satellite_batch._cc1 = satellite_batch._bstar * cc2
     cc3 = torch.where(satellite_batch._ecco > 1.0e-4,
                    -2.0 * coef * tsi * satellite_batch._j3oj2 * satellite_batch._no_unkozai * sinio / satellite_batch._ecco,
                    torch.tensor(0.0))
     satellite_batch._x1mth2 = 1.0 - cosio2
     satellite_batch._cc4 = 2.0 * satellite_batch._no_unkozai * coef1 * ao * omeosq * \
                        (satellite_batch._eta * (2.0 + 0.5 * etasq) + satellite_batch._ecco *
                            (0.5 + 2.0 * etasq) - satellite_batch._j2 * tsi / (ao * psisq) *
                            (-3.0 * satellite_batch._con41 * (1.0 - 2.0 * eeta + etasq *
                            (1.5 - 0.5 * eeta)) + 0.75 * satellite_batch._x1mth2 *
                            (2.0 * etasq - eeta * (1.0 + etasq)) * (2.0 * satellite_batch._argpo).cos()))
     satellite_batch._cc5 = 2.0 * coef1 * ao * omeosq * (1.0 + 2.75 *
                        (etasq + eeta) + eeta * etasq)
     cosio4 = cosio2 ** 2
     temp1 = 1.5 * satellite_batch._j2 * pinvsq * satellite_batch._no_unkozai
     temp2 = 0.5 * temp1 * satellite_batch._j2 * pinvsq
     temp3 = -0.46875 * satellite_batch._j4 * pinvsq * pinvsq * satellite_batch._no_unkozai
     satellite_batch._mdot = satellite_batch._no_unkozai + 0.5 * temp1 * rteosq * satellite_batch._con41 + 0.0625 * \
                            temp2 * rteosq * (13.0 - 78.0 * cosio2 + 137.0 * cosio4)
     satellite_batch._argpdot = (-0.5 * temp1 * con42 + 0.0625 * temp2 *
                                (7.0 - 114.0 * cosio2 + 395.0 * cosio4) +
                                temp3 * (3.0 - 36.0 * cosio2 + 49.0 * cosio4))
     xhdot1 = -temp1 * cosio
     satellite_batch._nodedot = xhdot1 + (0.5 * temp2 * (4.0 - 19.0 * cosio2) +
                                        2.0 * temp3 * (3.0 - 7.0 * cosio2)) * cosio
     xpidot = satellite_batch._argpdot + satellite_batch._nodedot
     satellite_batch._omgcof = satellite_batch._bstar * cc3 * satellite_batch._argpo.cos()
     satellite_batch._xmcof = torch.where(satellite_batch._ecco > 1.0e-4,
                                        -x2o3 * coef * satellite_batch._bstar / eeta,
                                        satellite_batch._xmcof)
     satellite_batch._nodecf = 3.5 * omeosq * xhdot1 * satellite_batch._cc1
     satellite_batch._t2cof = 1.5 * satellite_batch._cc1
     satellite_batch._xlcof = torch.where(torch.abs(cosio + 1.0) > 1.5e-12,
                                        -0.25 * satellite_batch._j3oj2 * sinio * (3.0 + 5.0 * cosio) / (1.0 + cosio),
                                        -0.25 * satellite_batch._j3oj2 * sinio * (3.0 + 5.0 * cosio) / temp4)
     satellite_batch._aycof = -0.5 * satellite_batch._j3oj2 * sinio
     delmotemp = 1.0 + satellite_batch._eta * satellite_batch._mo.cos()
     satellite_batch._delmo = delmotemp ** 3
     satellite_batch._sinmao = satellite_batch._mo.sin()
     satellite_batch._x7thm1 = 7.0 * cosio2 - 1.0

     # Deep space initialization
     deep_space_condition = (2 * np.pi / satellite_batch._no_unkozai >= 225.0)
     if torch.any(deep_space_condition):
         raise RuntimeError("Error: deep space propagation not supported (yet). One of the provided satellites has"
                        "an orbital period above 225 minutes. If you want to let us know you need it or you want to "
                        "contribute to implement it, open a PR or raise an issue at: https://github.com/esa/dSGP4.")

     isimp_condition = (satellite_batch._isimp != 1)
     if torch.any(isimp_condition):
         cc1sq = satellite_batch._cc1 ** 2
         d2 = 4.0 * ao * tsi * cc1sq
         temp = d2 * tsi * satellite_batch._cc1 / 3.0
         d3 = (17.0 * ao + sfour) * temp
         d4 = 0.5 * temp * ao * tsi * (221.0 * ao + 31.0 * sfour) * satellite_batch._cc1
         t3cof = d2 + 2.0 * cc1sq
         t4cof = 0.25 * (3.0 * d3 + satellite_batch._cc1 * (12.0 * d2 + 10.0 * cc1sq))
         t5cof = 0.2 * (3.0 * d4 + 12.0 * satellite_batch._cc1 * d3 +
                    6.0 * d2 ** 2 + 15.0 * cc1sq * (2.0 * d2 + cc1sq))
         satellite_batch._d2 = torch.where(isimp_condition, d2, satellite_batch._d2)
         satellite_batch._d3 = torch.where(isimp_condition, d3, satellite_batch._d3)
         satellite_batch._d4 = torch.where(isimp_condition, d4, satellite_batch._d4)
         satellite_batch._t3cof = torch.where(isimp_condition, t3cof, satellite_batch._t3cof)
         satellite_batch._t4cof = torch.where(isimp_condition, t4cof, satellite_batch._t4cof)
         satellite_batch._t5cof = torch.where(isimp_condition, t5cof, satellite_batch._t5cof)
     sgp4_batched(satellite_batch, torch.zeros((batch_size,)))

     satellite_batch._init = 'n'



__version__ = '1.0.0'

import torch
torch.set_default_dtype(torch.float64)
from .sgp4 import sgp4
from .initl import initl
from .sgp4init import sgp4init
from .sgp4init_batch import sgp4init_batch
from .newton_method import newton_method, update_TLE
from .sgp4_batched import sgp4_batched
from .util import propagate, initialize_tle, propagate_batch
from .plot import plot_orbit, plot_tles
from . import tle
from .tle import TLE



import torch
import numpy 
from .tle import TLE

def sgp4_batched(satellite_batch, tsince):
    """
    This function represents the batch SGP4 propagator. 
    It resembles `sgp4`, but accepts batches of TLEs.
    Having created the TLE object, and initialized the propagator (using `dsgp4.sgp4.sgp4init`), 
    one can use this method to propagate the TLE at future times. 
    The method returns the satellite position and velocity
    in km and km/s, respectively, after `tsince` minutes.

    Args:
        - satellite (``dsgp4.tle.TLE``): TLE batch object (with attributes that are N-dimensional tensors) 
        - tsince (``torch.tensor``): time to propagate, since the TLE epoch, in minutes (also an N-dimensional tensor)

    Returns:
        - batch_state (``torch.tensor``): a batch of 2x3 tensors, where the first row represents the spacecraft
                                    position (in km) and the second the spacecraft velocity (in km/s)
    """
    if not isinstance(satellite_batch, TLE):
        raise ValueError("satellite_batch should be a TLE object.")
    if not torch.is_tensor(tsince):
        raise ValueError("tsince must be a tensor.")
    if tsince.ndim!=1:
        raise ValueError("tsince should be a one dimensional tensor.")
    if len(tsince)!=len(satellite_batch._argpo):
        raise ValueError(f"in batch mode, tsince and satellite_batch shall have attributes of same length. Instead {len(tsince)} for time, and {len(satellite_batch._argpo)} for satellites' attributes found")
    if not hasattr(satellite_batch, '_radiusearthkm'):
        raise AttributeError('It looks like the satellite_batch has not been initialized. Please use the `initialize_tle` method or directly `sgp4init` to initialize the satellite_batch. Otherwise, if you are propagating, another option is to use `dsgp4.propagate` and pass `initialized=True` in the arguments.')
    
    batch_size = len(tsince)
    mrt = torch.zeros(batch_size)
    x2o3  = torch.tensor(2.0 / 3.0)

    vkmpersec    = torch.ones(batch_size)*(satellite_batch._radiusearthkm * satellite_batch._xke/60.0)
    #  sgp4 error flag
    satellite_batch._t    = tsince.clone()
    satellite_batch._error = torch.zeros(batch_size)

    # update for secular gravity and atmospheric drag
    xmdf    = satellite_batch._mo + satellite_batch._mdot * satellite_batch._t
    argpdf  = satellite_batch._argpo + satellite_batch._argpdot * satellite_batch._t
    nodedf  = satellite_batch._nodeo + satellite_batch._nodedot * satellite_batch._t
    argpm1   = argpdf
    mm1     = xmdf
    t2     = satellite_batch._t * satellite_batch._t
    nodem   = nodedf + satellite_batch._nodecf * t2
    tempa1   = 1.0 - satellite_batch._cc1 * satellite_batch._t
    tempe1   = satellite_batch._bstar * satellite_batch._cc4 * satellite_batch._t
    templ1   = satellite_batch._t2cof * t2


    delomg = satellite_batch._omgcof * satellite_batch._t

    delmtemp =  1.0 + satellite_batch._eta * xmdf.cos()
    delm   = satellite_batch._xmcof * \
                (delmtemp * delmtemp * delmtemp -
                satellite_batch._delmo)
    temp   = delomg + delm
    mm0     = xmdf + temp
    argpm0  = argpdf - temp
    t3     = t2 * satellite_batch._t
    t4     = t3 * satellite_batch._t
    tempa0  = tempa1 - satellite_batch._d2 * t2 - satellite_batch._d3 * t3 - \
                        satellite_batch._d4 * t4
    tempe0  = tempe1 + satellite_batch._bstar * satellite_batch._cc5 * (mm0.sin() -
                        satellite_batch._sinmao)
    templ0  = templ1 + satellite_batch._t3cof * t3 + t4 * (satellite_batch._t4cof +
                        satellite_batch._t * satellite_batch._t5cof)
    
    mm    = torch.where(satellite_batch._isimp==0,mm0,mm1)
    argpm = torch.where(satellite_batch._isimp==0,argpm0,argpm1)
    tempa = torch.where(satellite_batch._isimp==0,tempa0,tempa1)
    tempe = torch.where(satellite_batch._isimp==0,tempe0,tempe1)
    templ = torch.where(satellite_batch._isimp==0,templ0,templ1)

    nm    = satellite_batch._no_unkozai.clone()
    em    = satellite_batch._ecco.clone()
    inclm = satellite_batch._inclo.clone()

    satellite_batch._error = torch.full(nm.size(),2) * (nm <= 0.)

    am = torch.pow((satellite_batch._xke / nm),x2o3) * tempa * tempa
    nm = satellite_batch._xke / torch.pow(am, 1.5)
    em = em - tempe
    
    satellite_batch._error = torch.full(em.size(), 1) * ((em>=1.0) | (em<-0.001))

    em=torch.clamp(em, min=1.0e-6)
    mm     = mm + satellite_batch._no_unkozai * templ
    xlm    = mm + argpm + nodem
    emsq   = em * em
    temp   = 1.0 - emsq

    nodem = torch.fmod(nodem, torch.tensor(2*numpy.pi))

    argpm  = argpm % (2*numpy.pi)
    xlm    = xlm % (2*numpy.pi)
    mm     = (xlm - argpm - nodem) % (2*numpy.pi)

    satellite_batch._am = am.clone()
    satellite_batch._em = em.clone()
    satellite_batch._im = inclm.clone()
    satellite_batch._Om = nodem.clone()
    satellite_batch._om = argpm.clone()
    satellite_batch._mm = mm.clone()
    satellite_batch._nm = nm.clone()

    # compute extra mean quantities
    sinim = inclm.sin()
    cosim = inclm.cos()

    # add lunar-solar periodics
    ep     = em
    xincp  = inclm
    argpp  = argpm
    nodep  = nodem
    mp     = mm
    sinip  = sinim
    cosip  = cosim

    axnl = ep * argpp.cos()
    temp = 1.0 / (am * (1.0 - ep * ep))
    aynl = ep* argpp.sin() + temp * satellite_batch._aycof
    xl   = mp + argpp + nodep + temp * satellite_batch._xlcof * axnl

    #  solve kepler's equation 
    u    = (xl - nodep) % (2*numpy.pi)
    eo1  = u
    tem5 = torch.ones(tsince.size())

    for _ in range(10):
        coseo1=eo1.cos()
        sineo1=eo1.sin()
        tem5   = 1.0 - coseo1 * axnl - sineo1 * aynl
        tem5   = (u - aynl * coseo1 + axnl * sineo1 - eo1) / tem5
        tem5=torch.where(tem5>=0.95, 0.95, tem5)
        tem5=torch.where(tem5<=-0.95, -0.95, tem5)
        eo1    = eo1 + tem5

    # short period preliminary quantities
    ecose = axnl*coseo1 + aynl*sineo1
    esine = axnl*sineo1 - aynl*coseo1
    el2   = axnl*axnl + aynl*aynl
    pl    = am*(1.0-el2)
    satellite_batch._error=torch.where(satellite_batch._error==0.,torch.any(pl<0.)*4,1)

    rl     = am * (1.0 - ecose)
    rdotl  = am.sqrt() * esine/rl
    rvdotl = pl.sqrt() / rl
    betal  = (1.0 - el2).sqrt()
    temp   = esine / (1.0 + betal)
    sinu   = am / rl * (sineo1 - aynl - axnl * temp)
    cosu   = am / rl * (coseo1 - axnl + aynl * temp)
    su     = torch.atan2(sinu, cosu)
    sin2u  = (cosu + cosu) * sinu
    cos2u  = 1.0 - 2.0 * sinu * sinu
    temp   = 1.0 / pl
    temp1  = 0.5 * satellite_batch._j2 * temp
    temp2  = temp1 * temp

    mrt   = rl * (1.0 - 1.5 * temp2 * betal * satellite_batch._con41) + \
             0.5 * temp1 * satellite_batch._x1mth2 * cos2u
    su    = su - 0.25 * temp2 * satellite_batch._x7thm1 * sin2u
    xnode = nodep + 1.5 * temp2 * cosip * sin2u
    xinc  = xincp + 1.5 * temp2 * cosip * sinip * cos2u
    mvt   = rdotl - nm * temp1 * satellite_batch._x1mth2 * sin2u / satellite_batch._xke
    rvdot = rvdotl + nm * temp1 * (satellite_batch._x1mth2 * cos2u +
             1.5 * satellite_batch._con41) / satellite_batch._xke

    #  orientation vectors
    sinsu =  su.sin()
    cossu =  su.cos()
    snod  =  xnode.sin()
    cnod  =  xnode.cos()
    sini  =  xinc.sin()
    cosi  =  xinc.cos()
    xmx   = -snod * cosi
    xmy   =  cnod * cosi
    ux    =  xmx * sinsu + cnod * cossu
    uy    =  xmy * sinsu + snod * cossu
    uz    =  sini * sinsu
    vx    =  xmx * cossu - cnod * sinsu
    vy    =  xmy * cossu - snod * sinsu
    vz    =  sini * cossu

    # position and velocity (in km and km/sec)
    _mr = mrt * satellite_batch._radiusearthkm

    r = torch.stack((_mr * ux, _mr * uy, _mr * uz))
    v = torch.stack(((mvt * ux + rvdot * vx) * vkmpersec,
          (mvt * uy + rvdot * vy) * vkmpersec,
          (mvt * uz + rvdot * vz) * vkmpersec))

    satellite_batch._error=torch.where(satellite_batch._error==0.,torch.any(mrt<1.0)*6,satellite_batch._error)
    return torch.transpose(torch.stack((r.squeeze(),v.squeeze()),1),0,-1)#torch.cat((r.swapaxes(0,2),v.swapaxes(0,2)),1)#torch.stack(list(r)+list(v)).reshape(2,3)



import datetime
import dsgp4
import numpy as np
import random
import torch
import unittest

class UtilTestCase(unittest.TestCase):
    def test_newton_method_1(self):
        lines=[]
        lines.append('1 43437U 18100A   20143.90384230  .00041418  00000-0  10000-3 0 99968')
        lines.append('2 43437  97.8268 249.9127 0221000 123.9136 259.1144 15.12608579563539')
        my_tle=dsgp4.tle.TLE(lines)
        whichconst=dsgp4.util.get_gravity_constants("wgs-72")
        dsgp4.sgp4init(whichconst=whichconst,
                opsmode=my_tle._opsmode,
                satn=my_tle.satellite_catalog_number,
                epoch=(my_tle._jdsatepoch+my_tle._jdsatepochF)-2433281.5,
                xbstar=my_tle._bstar,
                xndot=my_tle._ndot,
                xnddot=my_tle._nddot,
                xecco=my_tle._ecco,
                xargpo=my_tle._argpo,
                xinclo=my_tle._inclo,
                xmo=my_tle._mo,
                xno_kozai=my_tle._no_kozai,
                xnodeo=my_tle._nodeo,
                satellite=my_tle)

        found_tle,_=dsgp4.newton_method(tle_0=my_tle, time_mjd=dsgp4.util.from_datetime_to_mjd(my_tle._epoch))
        self.assertEqual(my_tle.line1, found_tle.line1)
        self.assertEqual(my_tle.line2, found_tle.line2)
        #I now propagate for 1000 minutes and make sure that the found TLE w Newton
        #actually satisfies the tolerance
        tsince_2=1000.
        new_tol=1e-13
        target_state=dsgp4.sgp4(my_tle,tsince_2*torch.ones(1,1,requires_grad=True))
        time_mjd_0=dsgp4.util.from_datetime_to_mjd(my_tle._epoch)
        time_mjd=time_mjd_0+tsince_2/24/60
        found_tle_2,_=dsgp4.newton_method(tle_0=my_tle, time_mjd=time_mjd, new_tol=new_tol, target_state=target_state)
        found_state=dsgp4.sgp4(found_tle_2,torch.tensor((time_mjd-dsgp4.util.from_datetime_to_mjd(found_tle_2._epoch))*1440.))#(dsgp4.util.from_datetime_to_mjd(found_tle_2._epoch)-dsgp4.util.from_datetime_to_mjd(new_date))*1440.*torch.ones(1,1,requires_grad=True))
        self.assertTrue(torch.norm(found_state-target_state)<new_tol)



import numpy
import torch

from . import util

def initl(
       xke, j2,
       ecco,   epoch,  inclo,   no,
       method,
       opsmode,
       ):

     x2o3   = torch.tensor(2.0 / 3.0);

     eccsq  = ecco * ecco;
     omeosq = 1.0 - eccsq;
     rteosq = omeosq.sqrt();
     cosio  = inclo.cos();
     cosio2 = cosio * cosio;

     ak    = torch.pow(xke / no, x2o3);
     d1    = 0.75 * j2 * (3.0 * cosio2 - 1.0) / (rteosq * omeosq);
     del_  = d1 / (ak * ak);
     adel  = ak * (1.0 - del_ * del_ - del_ *
             (1.0 / 3.0 + 134.0 * del_ * del_ / 81.0));
     del_  = d1/(adel * adel);
     no    = no / (1.0 + del_);

     ao    = torch.pow(xke / no, x2o3);
     sinio = inclo.sin();
     po    = ao * omeosq;
     con42 = 1.0 - 5.0 * cosio2;
     con41 = -con42-cosio2-cosio2;
     ainv  = 1.0 / ao;
     posq  = po * po;
     rp    = ao * (1.0 - ecco);
     method = 'n';

     if opsmode == 'a':
         #  gst time
         ts70  = epoch - 7305.0;
         ds70 = torch.floor_divide(ts70 + 1.0e-8,1);
         tfrac = ts70 - ds70;
         #  find greenwich location at epoch
         c1    = torch.tensor(1.72027916940703639e-2);
         thgr70= torch.tensor(1.7321343856509374);
         fk5r  = torch.tensor(5.07551419432269442e-15);
         c1p2p = c1 + (2*numpy.pi);
         gsto  = (thgr70 + c1*ds70 + c1p2p*tfrac + ts70*ts70*fk5r) % (2*numpy.pi)
         if gsto < 0.0:
             gsto = gsto + (2*numpy.pi);

     else:
        gsto = util.gstime(epoch + 2433281.5);

     return (
       no,
       method,
       ainv,  ao,    con41,  con42, cosio,
       cosio2,eccsq, omeosq, posq,
       rp,    rteosq,sinio , gsto,
       )



import dsgp4
import sgp4
import sgp4.earth_gravity
import unittest


class UtilTestCase(unittest.TestCase):
    def test_earth_gravity_constants(self):
        out=dsgp4.util.get_gravity_constants("wgs-72")
        tumin,mu,radiusearthkm,xke,j2,j3,j4,j3oj2=out
        self.assertAlmostEqual(float(tumin),sgp4.earth_gravity.wgs72.tumin,places=10)
        self.assertAlmostEqual(float(mu),sgp4.earth_gravity.wgs72.mu,places=10)
        self.assertAlmostEqual(float(radiusearthkm),sgp4.earth_gravity.wgs72.radiusearthkm,places=10)
        self.assertAlmostEqual(float(xke),sgp4.earth_gravity.wgs72.xke,places=10)
        self.assertAlmostEqual(float(j2),sgp4.earth_gravity.wgs72.j2,places=10)
        self.assertAlmostEqual(float(j3),sgp4.earth_gravity.wgs72.j3,places=10)
        self.assertAlmostEqual(float(j4),sgp4.earth_gravity.wgs72.j4,places=10)
        self.assertAlmostEqual(float(j3oj2),sgp4.earth_gravity.wgs72.j3oj2,places=10)

        out=dsgp4.util.get_gravity_constants("wgs-72old")
        tumin,mu,radiusearthkm,xke,j2,j3,j4,j3oj2=out
        self.assertAlmostEqual(float(tumin),sgp4.earth_gravity.wgs72old.tumin,places=10)
        self.assertAlmostEqual(float(mu),sgp4.earth_gravity.wgs72old.mu,places=10)
        self.assertAlmostEqual(float(radiusearthkm),sgp4.earth_gravity.wgs72old.radiusearthkm,places=10)
        self.assertAlmostEqual(float(xke),sgp4.earth_gravity.wgs72old.xke,places=10)
        self.assertAlmostEqual(float(j2),sgp4.earth_gravity.wgs72old.j2,places=10)
        self.assertAlmostEqual(float(j3),sgp4.earth_gravity.wgs72old.j3,places=10)
        self.assertAlmostEqual(float(j4),sgp4.earth_gravity.wgs72old.j4,places=10)
        self.assertAlmostEqual(float(j3oj2),sgp4.earth_gravity.wgs72old.j3oj2,places=10)

        out=dsgp4.util.get_gravity_constants("wgs-84")
        tumin,mu,radiusearthkm,xke,j2,j3,j4,j3oj2=out
        self.assertAlmostEqual(float(tumin),sgp4.earth_gravity.wgs84.tumin,places=10)
        self.assertAlmostEqual(float(mu),sgp4.earth_gravity.wgs84.mu,places=10)
        self.assertAlmostEqual(float(radiusearthkm),sgp4.earth_gravity.wgs84.radiusearthkm,places=10)
        self.assertAlmostEqual(float(xke),sgp4.earth_gravity.wgs84.xke,places=10)
        self.assertAlmostEqual(float(j2),sgp4.earth_gravity.wgs84.j2,places=10)
        self.assertAlmostEqual(float(j3),sgp4.earth_gravity.wgs84.j3,places=10)
        self.assertAlmostEqual(float(j4),sgp4.earth_gravity.wgs84.j4,places=10)
        self.assertAlmostEqual(float(j3oj2),sgp4.earth_gravity.wgs84.j3oj2,places=10)



import numpy
import torch
from .initl import initl
from .sgp4 import sgp4

def sgp4init(
       whichconst,   opsmode,  satn, epoch,
       xbstar,   xndot,   xnddot,   xecco,   xargpo,
       xinclo,   xmo,   xno_kozai,
       xnodeo,  satellite,
       ):
     """
     This function initializes the sgp4 propagator.
     Args:
        - whichconst (``tuple``): this contains all the necessary constants (tumin, mu, radiusearthkm, xke,
                        j2, j3, j4, j3oj2)), according to the chosen gravitational model (wgs-72, wgs-84,
                        wgs-72old are three possible choices)
        - opsmode (``str``): mode of operation (possibilities: afspc or improved, 'a' and 'i', respectively)
        - satn (``str``): satellite catalog number
        - epoch (``torch.float``): TLE days since 1949 December 31 00:00 UT
        - xbstar (``torch.float``): drag coefficient (in 1/earth_radii)
        - xndot (``torch.float``): first derivative of mean motion (in revs/day**2)
        - xnddot (``torch.float``): second derivative mean motion (in revs/day**3)
        - xecco (``torch.float``): eccentricity
        - xargpo (``torch.float``): argument of perigee [rad]
        - xinclo (``torch.float``): inclination [rad]
        - xmo (``torch.float``): mean anomaly [rad]
        - xno_kozai (``torch.float``): mean motion (radians/minute)
        - xnodeo (``torch.float``): right ascension of the ascending node [rad]
        - satellite (``dsgp4.tle.TLE``): TLE object

    ..Note:
        If satellite._error is different than 0 once this routine is called, then the satellite propagation
        has had an error. In particular, the following error values hold for `satellite._error`:
        * `1` -> eccentricity `>=1.` or `<-0.001` or semi-major axis (in Earth radii) `<0.95`
        * `2` -> mean_motion `<0.`
        * `3` -> eccentricity `<0.` or `>1.`
        * `4` -> semi-latus rectum `<0.`
        * `5` -> epoch elements are sub-orbital
        * `6` -> satellite has decayed
     """
     temp4    =   torch.tensor(1.5e-12)

    #  ----------- set all near earth variables to zero ------------
     satellite._isimp   = torch.tensor(0);   satellite._method = 'n';               satellite._aycof    = torch.tensor(0.0);
     satellite._con41   = torch.tensor(0.0); satellite._cc1    = torch.tensor(0.0); satellite._cc4      = torch.tensor(0.0);
     satellite._cc5     = torch.tensor(0.0); satellite._d2     = torch.tensor(0.0); satellite._d3       = torch.tensor(0.0);
     satellite._d4      = torch.tensor(0.0); satellite._delmo  = torch.tensor(0.0); satellite._eta      = torch.tensor(0.0);
     satellite._argpdot = torch.tensor(0.0); satellite._omgcof = torch.tensor(0.0); satellite._sinmao   = torch.tensor(0.0);
     satellite._t       = torch.tensor(0.0); satellite._t2cof  = torch.tensor(0.0); satellite._t3cof    = torch.tensor(0.0);
     satellite._t4cof   = torch.tensor(0.0); satellite._t5cof  = torch.tensor(0.0); satellite._x1mth2   = torch.tensor(0.0);
     satellite._x7thm1  = torch.tensor(0.0); satellite._mdot   = torch.tensor(0.0); satellite._nodedot  = torch.tensor(0.0);
     satellite._xlcof   = torch.tensor(0.0); satellite._xmcof  = torch.tensor(0.0); satellite._nodecf   = torch.tensor(0.0);

     #  ------------------------ earth constants -----------------------
     #  sgp4fix identify constants and allow alternate values
     #  this is now the only call for the constants
     (satellite._tumin, satellite._mu, satellite._radiusearthkm, satellite._xke,
       satellite._j2, satellite._j3, satellite._j4, satellite._j3oj2) = whichconst

 	 # -------------------------------------------------------------------------

     satellite._error = torch.tensor(0)
     satellite._operationmode = opsmode
     satellite._satnum = satn

     satellite._bstar   = xbstar.clone()
    # sgp4fix allow additional parameters in the struct
     satellite._ndot    = xndot.clone()
     satellite._nddot   = xnddot.clone()
     satellite._ecco    = xecco.clone()
     satellite._argpo   = xargpo.clone()
     satellite._inclo   = xinclo.clone()
     satellite._mo	    = xmo.clone()
	# sgp4fix rename variables to clarify which mean motion is intended
     satellite._no_kozai= xno_kozai.clone()
     satellite._nodeo   = xnodeo.clone()

    # single averaged mean elements
     satellite._am = torch.tensor(0.0)
     satellite._em = torch.tensor(0.0)
     satellite._im = torch.tensor(0.0)
     satellite._Om = torch.tensor(0.0)
     satellite._mm = torch.tensor(0.0)
     satellite._nm = torch.tensor(0.0)

     ss     = 78.0 / satellite._radiusearthkm + 1.0

     qzms2ttemp = (120.0 - 78.0) / satellite._radiusearthkm
     qzms2t = qzms2ttemp * qzms2ttemp * qzms2ttemp * qzms2ttemp
     x2o3   =  2.0 / 3.0

     satellite._init = 'y'
     satellite._t	 = 0.0

    # sgp4fix remove satn as it is not needed in initl
     (
       satellite._no_unkozai,
       method,
       ainv,  ao,    satellite._con41,  con42, cosio,
       cosio2,eccsq, omeosq, posq,
       rp,    rteosq,sinio , satellite._gsto,
       ) = initl(
           satellite._xke, satellite._j2, satellite._ecco, epoch, satellite._inclo, satellite._no_kozai, satellite._method,
           satellite._operationmode
         )
     satellite._a    = torch.pow( satellite._no_unkozai*satellite._tumin , (-2.0/3.0) )
     satellite._alta = satellite._a*(1.0 + satellite._ecco) - 1.0
     satellite._altp = satellite._a*(1.0 - satellite._ecco) - 1.0

     if omeosq >= 0.0 or satellite._no_unkozai >= 0.0:

         satellite._isimp = 0
         if rp < 220.0 / satellite._radiusearthkm + 1.0:
             satellite._isimp = 1
         sfour  = ss
         qzms24 = qzms2t
         perige = (rp - 1.0) * satellite._radiusearthkm

         if perige < 156.0:
             sfour = perige - 78.0
             if perige < 98.0:
                 sfour = 20.0
             qzms24temp =  (120.0 - sfour) / satellite._radiusearthkm
             qzms24 = qzms24temp * qzms24temp * qzms24temp * qzms24temp
             sfour  = sfour / satellite._radiusearthkm + 1.0

         pinvsq = 1.0 / posq

         tsi  = 1.0 / (ao - sfour)
         satellite._eta  = ao * satellite._ecco * tsi
         etasq = satellite._eta * satellite._eta
         eeta  = satellite._ecco * satellite._eta
         psisq = torch.abs(1.0 - etasq)
         coef  = qzms24 * torch.pow(tsi, 4.0)
         coef1 = coef / torch.pow(psisq, 3.5)
         cc2   = coef1 * satellite._no_unkozai * (ao * (1.0 + 1.5 * etasq + eeta *
                        (4.0 + etasq)) + 0.375 * satellite._j2 * tsi / psisq * satellite._con41 *
                        (8.0 + 3.0 * etasq * (8.0 + etasq)))
         satellite._cc1   = satellite._bstar * cc2
         cc3   = 0.0
         if satellite._ecco > 1.0e-4:
             cc3 = -2.0 * coef * tsi * satellite._j3oj2 * satellite._no_unkozai * sinio / satellite._ecco
         satellite._x1mth2 = 1.0 - cosio2
         satellite._cc4    = 2.0* satellite._no_unkozai * coef1 * ao * omeosq * \
                           (satellite._eta * (2.0 + 0.5 * etasq) + satellite._ecco *
                           (0.5 + 2.0 * etasq) - satellite._j2 * tsi / (ao * psisq) *
                           (-3.0 * satellite._con41 * (1.0 - 2.0 * eeta + etasq *
                           (1.5 - 0.5 * eeta)) + 0.75 * satellite._x1mth2 *
                           (2.0 * etasq - eeta * (1.0 + etasq)) * (2.0 * satellite._argpo).cos()))
         satellite._cc5 = 2.0 * coef1 * ao * omeosq * (1.0 + 2.75 *
                        (etasq + eeta) + eeta * etasq)
         cosio4 = cosio2 * cosio2
         temp1  = 1.5 * satellite._j2 * pinvsq * satellite._no_unkozai
         temp2  = 0.5 * temp1 * satellite._j2 * pinvsq
         temp3  = -0.46875 * satellite._j4 * pinvsq * pinvsq * satellite._no_unkozai
         satellite._mdot     = satellite._no_unkozai + 0.5 * temp1 * rteosq * satellite._con41 + 0.0625 * \
                            temp2 * rteosq * (13.0 - 78.0 * cosio2 + 137.0 * cosio4)
         satellite._argpdot  = (-0.5 * temp1 * con42 + 0.0625 * temp2 *
                             (7.0 - 114.0 * cosio2 + 395.0 * cosio4) +
                             temp3 * (3.0 - 36.0 * cosio2 + 49.0 * cosio4))
         xhdot1            = -temp1 * cosio
         satellite._nodedot = xhdot1 + (0.5 * temp2 * (4.0 - 19.0 * cosio2) +
                              2.0 * temp3 * (3.0 - 7.0 * cosio2)) * cosio
         xpidot            =  satellite._argpdot+ satellite._nodedot
         satellite._omgcof   = satellite._bstar * cc3 * satellite._argpo.cos()
         if satellite._ecco > 1.0e-4:
             satellite._xmcof = -x2o3 * coef * satellite._bstar / eeta
         satellite._nodecf = 3.5 * omeosq * xhdot1 * satellite._cc1
         satellite._t2cof   = 1.5 * satellite._cc1
         if torch.abs(cosio+1.0) > 1.5e-12:
             satellite._xlcof = -0.25 * satellite._j3oj2 * sinio * (3.0 + 5.0 * cosio) / (1.0 + cosio)
         else:
             satellite._xlcof = -0.25 * satellite._j3oj2 * sinio * (3.0 + 5.0 * cosio) / temp4
         satellite._aycof   = -0.5 * satellite._j3oj2 * sinio
         delmotemp = 1.0 + satellite._eta * satellite._mo.cos()
         satellite._delmo   = delmotemp * delmotemp * delmotemp
         satellite._sinmao  = satellite._mo.sin()
         satellite._x7thm1  = 7.0 * cosio2 - 1.0

         #  --------------- deep space initialization -------------
         if 2*numpy.pi / satellite._no_unkozai >= 225.0:
             raise RuntimeError("Error: deep space propagation not supported (yet). The provided satellite has\
             an orbital period above 225 minutes. If you want to let us know you need it or you want to \
             contribute to implement it, open a PR or raise an issue at: https://github.com/esa/dSGP4.")

         if satellite._isimp != 1:
           cc1sq          = satellite._cc1 * satellite._cc1
           satellite._d2    = 4.0 * ao * tsi * cc1sq
           temp           = satellite._d2 * tsi * satellite._cc1 / 3.0
           satellite._d3    = (17.0 * ao + sfour) * temp
           satellite._d4    = 0.5 * temp * ao * tsi * (221.0 * ao + 31.0 * sfour) * \
                            satellite._cc1
           satellite._t3cof = satellite._d2 + 2.0 * cc1sq
           satellite._t4cof = 0.25 * (3.0 * satellite._d3 + satellite._cc1 *
                            (12.0 * satellite._d2 + 10.0 * cc1sq))
           satellite._t5cof = 0.2 * (3.0 * satellite._d4 +
                            12.0 * satellite._cc1 * satellite._d3 +
                            6.0 * satellite._d2 * satellite._d2 +
                            15.0 * cc1sq * (2.0 * satellite._d2 + cc1sq))
     sgp4(satellite, torch.zeros(1,1));

     satellite._init = 'n'


import datetime
import numpy as np
import torch
import unittest

import dsgp4.tle

class UtilTestCase(unittest.TestCase):
    def test_tle(self):
        data = dict(
            satellite_catalog_number=43437,
            classification='U',
            international_designator='18100A',
            epoch_year=2020,
            epoch_days=143.90384230,
            ephemeris_type=0,
            element_number=9996,
            revolution_number_at_epoch=56353,
            mean_motion=0.0011,
            mean_motion_first_derivative=6.9722e-13,
            mean_motion_second_derivative=.0,
            eccentricity=0.0221,
            inclination=1.7074,
            argument_of_perigee=2.1627,
            raan=4.3618,
            mean_anomaly=4.5224,
            b_star=0.0001)

        tle = dsgp4.tle.TLE(data)
        line1, line2 = tle.line1, tle.line2
        line1Correct = '1 43437U 18100A   20143.90384230  .00041418  00000-0  10000-3 0 99968'
        line2Correct = '2 43437  97.8268 249.9127 0221000 123.9136 259.1144 15.12608579563539'

        self.assertEqual(line1Correct, line1)
        self.assertEqual(line2Correct, line2)

        data = dict(
            satellite_catalog_number=43437,
            classification='U',
            international_designator='18100A',
            epoch_year=2020,
            epoch_days=143.90384230,
            ephemeris_type=0,
            element_number=9996,
            revolution_number_at_epoch=56353,
            mean_motion=0.0010,
            mean_motion_first_derivative=1.6261e-13,
            mean_motion_second_derivative=1e-5,
            eccentricity=0.0274,
            inclination=1.5086,
            argument_of_perigee=4.9757,
            raan=3.8719,
            mean_anomaly=6.0775,
            b_star=-0.0010)

        tle = dsgp4.tle.TLE(data)
        line1, line2 = tle.line1, tle.line2
        line1Correct = '1 43437U 18100A   20143.90384230  .00009660  17108+9 -10000-2 0 99966'
        line2Correct = '2 43437  86.4364 221.8435 0274000 285.0866 348.2151 13.75098708563531'

        self.assertEqual(line1Correct, line1)
        self.assertEqual(line2Correct, line2)

    def test_read_tle(self):
        #Sentinel-1A:
        tle_line1 = '1 39634U 14016A   22059.08188563  .00000057  00000+0  21846-4 0  9990'
        tle_line2 = '2 39634  98.1819  68.1874 0001341  82.4703 277.6657 14.59199732421074'
        tle = dsgp4.tle.TLE([tle_line1, tle_line2])
        datetime_object = datetime.datetime.strptime(tle['date_string'], '%Y-%m-%d %H:%M:%S.%f')

        #Now the values to validate:
        satellite_catalog_number = 39634
        classification = 'U'
        international_designator = '14016A'
        mean_motion_first_derivative = 9.595270771959568e-16
        mean_motion_second_derivative = 0.0
        b_star = 2.1846e-05
        year = 2022
        month = 2
        day = 28
        hour = 1
        minute = 57
        second = 54
        microsecond = 918432
        date_string = '2022-02-28 01:57:54.918432'
        ephemeris_type = 0
        element_number = 999
        inclination = 1.7135974208638207
        raan = 1.1900946383743813
        eccentricity = 0.0001341
        argument_of_perigee = 1.4393782701074795
        mean_anomaly = 4.8461806848548195
        mean_motion = 0.0010611599903174525
        revolution_number_at_epoch = 42107

        self.assertEqual(tle['international_designator'], international_designator)
        self.assertEqual(tle['classification'], classification)
        self.assertEqual(tle['satellite_catalog_number'], satellite_catalog_number)
        self.assertEqual(datetime_object.year, year)
        self.assertEqual(datetime_object.month, month)
        self.assertEqual(datetime_object.day, day)
        self.assertEqual(datetime_object.hour, hour)
        self.assertEqual(datetime_object.minute, minute)
        self.assertEqual(datetime_object.second, second)
        self.assertEqual(datetime_object.microsecond, microsecond)
        self.assertEqual(tle['date_string'], date_string)
        self.assertEqual(tle['ephemeris_type'], ephemeris_type)
        self.assertEqual(tle['element_number'], element_number)
        self.assertEqual(tle['revolution_number_at_epoch'], revolution_number_at_epoch)
        self.assertAlmostEqual(tle['b_star'], b_star, places = 10)
        self.assertAlmostEqual(tle['mean_motion_first_derivative'], mean_motion_first_derivative, places = 10)
        self.assertAlmostEqual(tle['mean_motion_second_derivative'], mean_motion_second_derivative, places = 10)
        self.assertAlmostEqual(tle['inclination'], inclination, places = 10)
        self.assertAlmostEqual(tle['raan'], raan, places = 10)
        self.assertAlmostEqual(tle['eccentricity'], eccentricity, places = 10)
        self.assertAlmostEqual(tle['argument_of_perigee'], argument_of_perigee, places = 10)
        self.assertAlmostEqual(tle['mean_anomaly'], mean_anomaly, places = 10)
        self.assertAlmostEqual(tle['mean_motion'], mean_motion, places = 10)

    def test_read_tle_make_tle(self):

        #we now test that from a TLE, I can read the elements and construct a new one:
        lines_in = ['1 00046U 60007B   22054.60355309  .00000135  00000-0  49836-4 0  9990',
                 '2 00046  66.6904 181.4288 0206178  29.9665 331.3056 14.49532408228403']

        tle = dsgp4.tle.TLE(lines_in)

        datetime_object = datetime.datetime.strptime(tle['date_string'], '%Y-%m-%d %H:%M:%S.%f')
        epoch_year = datetime_object.year
        epoch_days = dsgp4.util.from_datetime_to_fractional_day(datetime_object)

        tle2 = dsgp4.tle.TLE(data = tle._data)
        lines_out = [tle2.line1, tle2.line2]

        self.assertEqual(lines_in[0], lines_out[0])
        self.assertEqual(lines_in[1], lines_out[1])

    def test_load_tles(self):
        tle_1 = ['1 00046U 60007B   22054.60355309  .00000135  00000-0  49836-4 0  9990',
                 '2 00046  66.6904 181.4288 0206178  29.9665 331.3056 14.49532408228403']
        lines_1, data_1 = dsgp4.tle.load_from_lines(tle_1)
        lines_2, data_2 = dsgp4.tle.load_from_data(data_1)
        self.assertEqual(tle_1, lines_1)
        self.assertEqual(lines_1, lines_2)
        self.assertEqual(data_1, data_2)
        
    def test_tle_from_vallado(self):
        #this test is taken from page 107 of Dr. David Vallado's book: Fundamentals of Astrodynamics and Applications, 4th Edition
        tle_lines=['1 16609U 86017A   93352.53502934  .00007889  00000-0  10529-3 0   342',
                   '2 16609  51.6190  13.3340 0005770 102.5680 257.5950 15.59114070447869']
        tle=dsgp4.tle.TLE(tle_lines)
        #TLE elements:
        xpdotp   =  1440.0 / (2.0 *np.pi);
        self.assertAlmostEqual(float(tle._bstar),0.00010529, places = 8)
        self.assertAlmostEqual(float(tle._ndot)*(xpdotp*1440.0),7.889*1e-5, places = 8)
        self.assertAlmostEqual(float(tle._nddot),0.0, places = 8)
        self.assertAlmostEqual(float(tle._ecco), 0.0005770, places = 8)
        self.assertAlmostEqual(float(tle._nodeo)*180.0/np.pi, 13.3340, places = 8)
        self.assertAlmostEqual(float(tle._inclo)*180.0/np.pi, 51.6190, places = 8)
        self.assertAlmostEqual(float(tle._mo)*180.0/np.pi, 257.5950, places = 8)
        self.assertAlmostEqual(float(tle._argpo)*180.0/np.pi, 102.5680, places = 8)
        self.assertAlmostEqual(round(float(tle._no_kozai)*xpdotp,8), 15.59114070, places = 8)

        #epoch:
        self.assertAlmostEqual(tle._epoch.month,12, places = 8)
        self.assertAlmostEqual(tle._epoch.year,1993, places = 8)
        self.assertAlmostEqual(tle._epoch.day,18, places = 8)
        self.assertAlmostEqual(tle._epoch.hour,12, places = 8)
        self.assertAlmostEqual(tle._epoch.minute,50, places = 8)
        self.assertAlmostEqual(tle._epoch.second,26, places = 8)
        self.assertAlmostEqual(round(tle._epoch.microsecond*1e-6,4),0.5350, places = 8)



import datetime
import numpy as np
import copy
import torch

from . import util
_, MU_EARTH, _, _, _, _, _, _=util.get_gravity_constants('wgs-84')
MU_EARTH = MU_EARTH*1e9

# This function is from python-sgp4 released under MIT License, (c) 2012–2016 Brandon Rhodes
def compute_checksum(line):
    """
    This function takes a TLE line in the form of a string and computes the checksum.

    Args:
        line (`str`): TLE line

    Returns:
        `int`: checksum
    """
    return sum((int(c) if c.isdigit() else c == '-') for c in line[0:68]) % 10


# This function is from python-sgp4 released under MIT License, (c) 2012–2016 Brandon Rhodes
def read_satellite_catalog_number(string):
    """
    This function takes a string corresponding to part of a TLE line, and returns
    the corresponding satellite catalog number.

    Args:
        string (`str`): string line
    
    Returns:
        `int`: satellite catalog number
    """
    if not string[0].isalpha():
        return int(string)
    character = string[0]
    n = ord(character) - ord('A') + 10
    n -= character > 'I'
    n -= character > 'O'
    return n * 10000 + int(string[1:])


# Parts of this function is based on python-sgp4 released under MIT License, (c) 2012–2016 Brandon Rhodes
def load_from_lines(lines, opsmode='i'):
    """
    This function takes a TLE as a list of strings and returns both itself and its dictionary representation.
    Args:
        lines (`list`): TLE data in the form of a list
        opsmode (`str`): operation mode, either 'i' or 'a'
    Returns:
        `list`: TLE data in the form of a list
        `dict`: TLE data in the form of a dictionary
    """
    if isinstance(lines, str):
        lines = util.get_non_empty_lines(lines)
    elif isinstance(lines, list):
        if not isinstance(lines[0], str):
            raise ValueError('Expecting a list of strings')
    else:
        raise ValueError('Expecting either a string or a list of strings')

    if len(lines) == 2:
        line0, line1, line2 = None, lines[0], lines[1]
    elif len(lines) == 3:
        line0, line1, line2 = lines[0], lines[1], lines[2]
    else:
        raise ValueError('Expecting a string with two or three lines')
    #for SGP4:
    xpdotp   =  1440.0 / (2.0 *np.pi);
    # if type(line1)!= str or type(line2)!=str:
        # raise ValueError('Input shall be a list of two string elements (1st and 2nd line).')
    #we initialize the tle dictionary:
    data = {}
    #we remove any trailing character:
    line = line1.rstrip()
    if (len(line) >= 64 and
        line.startswith('1 ') and
        line[8] == ' ' and
        line[23] == '.' and
        line[32] == ' ' and
        line[34] == '.' and
        line[43] == ' ' and
        line[52] == ' ' and
        line[61] == ' ' and
        line[63] == ' '):

        data['satellite_catalog_number'] = read_satellite_catalog_number(line1[2:7])
        data['classification']           = line[7] or 'U'
        data['international_designator'] = line[9:17].rstrip()
        two_digit_year=int(line[18:20])
        if two_digit_year<57:
            year = two_digit_year + 2000#int('19'+line[18:20])
        else:
            year = two_digit_year + 1900
        epochdays = float(line[20:32])
        date_datetime = datetime.datetime(year-1, 12, 31, 0, 0, 0, 0)+datetime.timedelta(days = epochdays)
        date_string = date_datetime.strftime(format = '%Y-%m-%d %H:%M:%S.%f')
        data['epoch_year'] = date_datetime.year
        data['epoch_days'] = epochdays
        data['date_string'] = date_string
        data['date_mjd'] = util.from_datetime_to_mjd(util.from_string_to_datetime(date_string))
        data['mean_motion_first_derivative'] = float(line[33:43])*np.pi/1.86624e9
        data['mean_motion_second_derivative'] = float(line[44]+'.'+line[45:50])*pow(10, int(line[50:52]))*np.pi/5.3747712e13
        data['b_star'] = float(line[53]+'.'+line[54:59])*pow(10, int(line[59:61]))
        data['ephemeris_type'] = int(line[62])
        data['element_number'] = int(line[64:68])
        data['line1'] = line
        #for SGP4:
        data['_epochdays'] = epochdays
        data['_bstar'] = torch.tensor(float(line[53]+'.'+line[54:59])*pow(10, int(line[59:61])))
        data['_ndot'] = torch.tensor(float(line[33:43])/(xpdotp*1440.0))
        data['_nddot']= torch.tensor(float(line[44] + '.' + line[45:50])/(xpdotp*1440.0*1440))

    else:
        raise ValueError('First line not compatible with TLE format.')

    line = line2.rstrip()

    if (len(line) >= 69 and
        line.startswith('2 ') and
        line[7] == ' ' and
        line[11] == '.' and
        line[16] == ' ' and
        line[20] == '.' and
        line[25] == ' ' and
        line[33] == ' ' and
        line[37] == '.' and
        line[42] == ' ' and
        line[46] == '.' and
        line[51] == ' '):

        if data['satellite_catalog_number'] != read_satellite_catalog_number(line[2:7]):
            raise ValueError('Satellite catalog numbers of line1 and line2 do not match.')

        data['inclination'] = np.deg2rad(float(line[8:16]))
        data['raan'] = np.deg2rad(float(line[17:25]))
        data['eccentricity'] = float('0.'+line[26:33].replace(' ', '0'))
        data['argument_of_perigee'] = np.deg2rad(float(line[34:42]))
        data['mean_anomaly'] = np.deg2rad(float(line[43:51]))
        data['mean_motion'] = float(line[52:63])*np.pi/43200.0
        data['revolution_number_at_epoch'] = int(line[63:68])
        data['line2'] = line
        #for SGP4:
        data['_inclo']=torch.tensor(np.deg2rad(float(line[8:16])))
        data['_nodeo']=torch.tensor(np.deg2rad(float(line[17:25])))
        data['_ecco'] =torch.tensor(float('0.'+line[26:33].replace(' ', '0')))
        data['_argpo']=torch.tensor(np.deg2rad(float(line[34:42])))
        data['_mo']=torch.tensor(np.deg2rad(float(line[43:51])))
        data['_no_kozai'] =torch.tensor(float(line[52:63]) / xpdotp);

    else:
        raise ValueError('Second line not compatible with TLE format.')

    mon,day,hr,minute,sec = util.days2mdhms(year, epochdays);
    sec_whole, sec_fraction = divmod(sec, 1.0)
    data['_epochyr'] = torch.tensor(year)
    data['_jdsatepoch'] = torch.tensor(util.jday(year,mon,day,hr,minute,sec)[0]);
    data['_jdsatepochF'] = torch.tensor(util.jday(year,mon,day,hr,minute,sec)[1]);

    #I also add the semi-major axis:
    data['semi_major_axis'] = (MU_EARTH/(data['mean_motion']**2))**(1.0/3.0)

    try:
        data['_epoch'] = datetime.datetime(year, mon, day, hr, minute, int(sec_whole),
                                int(sec_fraction * 1000000.0 // 1.0))
    except ValueError:
        # Sometimes a TLE says something like "2019 + 366.82137887 days"
        # which would be December 32nd which causes a ValueError.
        year, mon, day, hr, minute, sec = util.invjday(data['_jdsatepoch'])
        data['_epoch'] = datetime.datetime(year, mon, day, hr, minute, int(sec_whole),
                                int(sec_fraction * 1000000.0 // 1.0))
    data['_opsmode']=opsmode
    # Process optional line zero
    if line0 is not None:
        data['line0'] = line0
        if line0.startswith('0 '):
            line0 = line0[2:]
        data['name'] = line0.strip()
    return lines, data


# Parts of this function is based on python-sgp4 released under MIT License, (c) 2012–2016 Brandon Rhodes
def load_from_data(data, opsmode='i'):
    """
    This function takes a TLE as a dictionary and returns both itself and its representation as a list of strings.

    Args:
        data (`dict`): TLE data in the form of a dictionary
        opsmode (`str`): 'i' for improved, 's' for simplified

    Returns:
        `list`: TLE data in the form of a list
        `dict`: TLE data in the form of a dictionary
    """
    #for SGP4:
    xpdotp   =  1440.0 / (2.0 *np.pi);

    data['raan']=data['raan']%(2*np.pi)
    data['argument_of_perigee']=data['argument_of_perigee']%(2*np.pi)
    data['mean_anomaly']=data['mean_anomaly']%(2*np.pi)
    line1 = ['1 ']
    line1.append(str(data['satellite_catalog_number']).zfill(5)[:5])
    line1.append(str(data['classification'])[0] + ' ')
    line1.append(str(data['international_designator']).ljust(8, ' ')[:8] + ' ')
    line1.append(str(data['epoch_year'])[-2:].zfill(2) + '{:012.8f}'.format(data['epoch_days']) + ' ')
    line1.append('{0: 8.8f}'.format(data['mean_motion_first_derivative'] * (1.86624e9 / np.pi)).replace('0', '', 1) + ' ')
    line1.append('{0: 4.4e}'.format((data['mean_motion_second_derivative'] * (5.3747712e13 / np.pi)) * 10).replace(".", '').replace('e+00', '-0').replace('e-0', '-').replace('e+0', '+') + ' ')
    line1.append('{0: 4.4e}'.format(data['b_star'] * 10).replace('.', '').replace('e+0', '+').replace('e-0', '-') + ' ')
    line1.append('{} '.format(data['ephemeris_type']) + str(data['element_number']).rjust(4, ' '))
    line1 = ''.join(line1)
    line1 += str(compute_checksum(line1))
    two_digit_year=int(line1[18:20])
    if two_digit_year<57:
        year = two_digit_year + 2000#int('19'+line1[18:20])
    else:
        year = two_digit_year + 1900#int('20'+line1[18:20])
    epochdays = float(line1[20:32])
    data['_epochdays'] = epochdays
    date_datetime = datetime.datetime(year-1, 12, 31, 0, 0, 0, 0)+datetime.timedelta(days = epochdays)
    date_string = date_datetime.strftime(format = '%Y-%m-%d %H:%M:%S.%f')
    data['epoch_year'] = date_datetime.year
    data['epoch_days'] = epochdays
    data['date_string'] = date_string
    #for SGP4:
    data['_bstar'] = torch.tensor(float(line1[53]+'.'+line1[54:59])*pow(10, int(line1[59:61])))
    data['_ndot'] = torch.tensor(float(line1[33:43])/(xpdotp*1440.0))
    data['_nddot']= torch.tensor(float(line1[44] + '.' + line1[45:50])/(xpdotp*1440.0*1440))

    line2 = ['2 ']
    line2.append(str(data['satellite_catalog_number']).zfill(5)[:5] + ' ')
    line2.append('{0:8.4f}'.format(data['inclination'] * (180 / np.pi)).rjust(8, ' ') + ' ')
    line2.append('{0:8.4f}'.format(data['raan'] * (180 / np.pi)).rjust(8, ' ') + ' ')
    line2.append(str(round(float(data['eccentricity']) * 1e7)).rjust(7, '0')[:7] + ' ')
    line2.append('{0:8.4f}'.format(data['argument_of_perigee'] * (180 / np.pi)).rjust(8, ' ') + ' ')
    line2.append('{0:8.4f}'.format(data['mean_anomaly'] * (180 / np.pi)).rjust(8, ' ') + ' ')
    line2.append('{0:11.8f}'.format(data['mean_motion'] * 43200.0 / np.pi).rjust(8, ' '))
    line2.append(str(data['revolution_number_at_epoch']).rjust(5))
    line2 = ''.join(line2)
    line2 += str(compute_checksum(line2))
    #for SGP4:
    data['_inclo']=torch.tensor(np.deg2rad(float(line2[8:16])))
    data['_nodeo']=torch.tensor(np.deg2rad(float(line2[17:25])))
    data['_ecco'] =torch.tensor(float('0.'+line2[26:33].replace(' ', '0')))
    data['_argpo']=torch.tensor(np.deg2rad(float(line2[34:42])))
    data['_mo']=torch.tensor(np.deg2rad(float(line2[43:51])))
    data['_no_kozai'] =torch.tensor(float(line2[52:63]) / xpdotp);

    if len(line1) != 69:
        raise RuntimeError('TLE line 1 has unexpected length ({})'.format(len(line1)))
    if len(line2) != 69:
        raise RuntimeError('TLE line 2 has unexpected length ({})'.format(len(line2)))

    data['line0'] = None
    data['line1'] = line1
    data['line2'] = line2
    date_datetime = datetime.datetime(data['epoch_year']-1, 12, 31, 0, 0, 0, 0)+datetime.timedelta(days = data['epoch_days'])
    date_string = date_datetime.strftime(format = '%Y-%m-%d %H:%M:%S.%f')
    data['date_string'] = date_string
    data['date_mjd'] = util.from_datetime_to_mjd(util.from_string_to_datetime(date_string))
    data['semi_major_axis'] = (MU_EARTH/(data['mean_motion']**2))**(1.0/3.0)
    lines = [line1, line2]
    mon,day,hr,minute,sec = util.days2mdhms(year, epochdays);
    sec_whole, sec_fraction = divmod(sec, 1.0)
    data['_epochyr'] = torch.tensor(year)
    data['_jdsatepoch'] = torch.tensor(util.jday(year,mon,day,hr,minute,sec)[0]);
    data['_jdsatepochF'] = torch.tensor(util.jday(year,mon,day,hr,minute,sec)[1]);

    #I also add the semi-major axis:
    data['semi_major_axis'] = (MU_EARTH/(data['mean_motion']**2))**(1.0/3.0)
    try:
        data['_epoch'] = datetime.datetime(year, mon, day, hr, minute, int(sec_whole),
                                int(sec_fraction * 1000000.0 // 1.0))
    except ValueError:
        # Sometimes a TLE says something like "2019 + 366.82137887 days"
        # which would be December 32nd which causes a ValueError.
        year, mon, day, hr, minute, sec = util.invjday(data['_jdsatepoch'])
        data['_epoch'] = datetime.datetime(year, mon, day, hr, minute, int(sec_whole),
                                int(sec_fraction * 1000000.0 // 1.0))
    data['_opsmode']=opsmode
    if 'name' in data:
        data['line0'] = '0 '+data['name']
    return lines, data


def load(file_name):
    """
    This function takes a file name that contains TLE (either with names or without),
    and returns a list of TLE objects.

    Args:
        file_name (`str`): TLEs file name

    Returns:
        `list`: list of `dsgp4.tle.TLE` objects
    """
    with open(file_name) as f:
        lines = util.get_non_empty_lines(f.read())

    i = 0
    tles = []
    while True:
        line = lines[i]
        if not (line.startswith('1 ') or line.startswith('2 ')):
            # This is a line0 (extra name line)
            tle = TLE(lines[i:i+3])
            i += 3
        else:
            # This is assumed to be line1 (first line of a TLE)
            tle = TLE(lines[i:i+2])
            i += 2
        tles.append(tle)
        if i == len(lines):
            break
    return tles


class TLE():
    """
    This class constructs a TLE object from either a list of strings (that make up the TLE)
    or from a dictionary.

    Args:
        tle (`list` or `dict`): list of strings (that make up the TLE) or dictionary

    Returns:
        `dsgp4.tle.TLE` object
    """
    def __init__(self, data):
        if isinstance(data, list) or isinstance(data, str):
            self._lines, self._data = load_from_lines(data)
        elif isinstance(data, dict):
            self._lines, self._data = load_from_data(data)
        else:
            raise RuntimeError('Expecting a string of TLE, list of strings with TLE lines, or a dictionary of TLE data.')

    def copy(self):
        """
        This function returns a copy of the TLE object.

        Returns:
            `dsgp4.tle.TLE` object
        """
        d = copy.deepcopy(self._data)
        return TLE(d)

    def set_time(self, date_mjd):
        """
        This function sets the epoch of the TLE to the given date.

        Args:
            date_mjd (`float`): date in MJD
        """
        d = copy.deepcopy(self._data)
        d['epoch_year'] = util.from_mjd_to_datetime(date_mjd).year
        d['epoch_days'] = util.from_mjd_to_epoch_days_after_1_jan(date_mjd)
        mon,day,hr,minute,sec = util.days2mdhms(d['epoch_year'], d['epoch_days']);
        sec_whole, sec_fraction = divmod(sec, 1.0)
        d['_epochyr'] = torch.tensor(d['epoch_year'])
        d['_jdsatepoch'] = torch.tensor(util.jday(d['epoch_year'],mon,day,hr,minute,sec)[0]);
        d['_jdsatepochF'] = torch.tensor(util.jday(d['epoch_year'],mon,day,hr,minute,sec)[1]);
        tle = TLE(d)
        self._data = tle._data
        self._lines = tle._lines

    def update(self, tle_data):
        """
        This function updates the TLE object with the given data.

        Args:
            tle_data (`dict`): dictionary of TLE data
        """
        d = copy.deepcopy(self._data)
        for k, v in tle_data.items():
            d[k] = v
        tle = TLE(d)
        self._mo=tle['_mo']
        self._bstar=tle['_bstar']
        self._ndot=tle['_ndot']
        self._nddot=tle['_nddot']
        self._ecco= tle['_ecco']
        self._argpo=tle['_argpo']
        self._inclo=tle['_inclo']
        self._no_kozai=tle['_no_kozai']
        self._nodeo=tle['_nodeo']
        self._data = tle._data
        self._lines = tle._lines

    def perigee_alt(self, R_eq = 6378135.0):
        """
        This function returns the perigee altitude of a given TLE.

        Args:
            R_eq (`float`): equatorial radius of the Earth [m]

        Returns:
            `float`: perigee altitude
        """
        return self.semi_major_axis*(1-self.eccentricity)-R_eq

    def apogee_alt(self, R_eq = 6378135.0):
        """
        This function returns the apogee altitude of a given TLE.

        Args:
            R_eq (`float`): equatorial radius of the Earth [m]

        Returns:
            `float`: apogee altitude
        """
        return self.semi_major_axis*(1+self.eccentricity)-R_eq

    def __repr__(self):
        return 'TLE(\n{}\n)'.format('\n'.join(self._lines))

    def __getitem__(self, index):
        return self._data[index]

    def __getattr__(self, attr):
        if attr == '_data' or attr == '_lines':
            return super().__getattr__(attr)
        elif attr in self._data.keys():
            return self[attr]
        else:
            return super().__getattr__(attr)




import os
import sys
from setuptools import setup, find_packages
PACKAGE_NAME = 'dsgp4'
MINIMUM_PYTHON_VERSION = 3, 5

with open('README.md', 'r') as f:
    long_description = f.read()

def check_python_version():
    """Exit when the Python version is too low."""
    if sys.version_info < MINIMUM_PYTHON_VERSION:
        sys.exit("Python {}.{}+ is required.".format(*MINIMUM_PYTHON_VERSION))


def read_package_variable(key):
    """Read the value of a variable from the package without importing."""
    module_path = os.path.join(PACKAGE_NAME, '__init__.py')
    with open(module_path) as module:
        for line in module:
            parts = line.strip().split(' ')
            if parts and parts[0] == key:
                return parts[-1].strip("'")
    assert 0, "'{0}' not found in '{1}'".format(key, module_path)


check_python_version()
setup(
    name='dsgp4',
    version=read_package_variable('__version__'),
    description='Differentiable SGP4 prototype',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Acciarini & Baydin',
    author_email='giacomo.acciarini@gmail.com',
    packages=find_packages(),
    url='https://github.com/esa/dSGP4',
    install_requires=['numpy', 'torch', 'matplotlib'],
    extras_require={'dev': ['pytest', 'coverage', 'pytest-xdist', 'sgp4>=2.21']},
    classifiers=['License :: OSI Approved :: GNU General Public License v3 (GPLv3)', 'Programming Language :: Python :: 3']
)



import datetime
import numpy as np
import torch

def get_gravity_constants(gravity_constant_name):
    if gravity_constant_name == 'wgs-72old':
        mu     = 398600.79964        #  in km3 / s2
        radiusearthkm = 6378.135     #  km
        xke    = 0.0743669161
        tumin  = 1.0 / xke
        j2     =   0.001082616
        j3     =  -0.00000253881
        j4     =  -0.00000165597
        j3oj2  =  j3 / j2
    elif gravity_constant_name == 'wgs-72':
       mu     = 398600.8            #  in km3 / s2
       radiusearthkm = 6378.135     #  km
       xke    = 60.0 / np.sqrt(radiusearthkm*radiusearthkm*radiusearthkm/mu)
       tumin  = 1.0 / xke
       j2     =   0.001082616
       j3     =  -0.00000253881
       j4     =  -0.00000165597
       j3oj2  =  j3 / j2
    elif gravity_constant_name=="wgs-84":
       mu     = 398600.5            #  in km3 / s2
       radiusearthkm = 6378.137     #  km
       xke    = 60.0 / np.sqrt(radiusearthkm*radiusearthkm*radiusearthkm/mu)
       tumin  = 1.0 / xke
       j2     =   0.00108262998905
       j3     =  -0.00000253215306
       j4     =  -0.00000161098761
       j3oj2  =  j3 / j2
    else:
       raise RuntimeError("Supported gravity constant names: wgs-72, wgs-84, wgs-72old while "+gravity_constant_name+" was provided")

    return torch.tensor(tumin), torch.tensor(mu), torch.tensor(radiusearthkm), torch.tensor(xke), torch.tensor(j2), torch.tensor(j3), torch.tensor(j4), torch.tensor(j3oj2)

def propagate_batch(tles, tsinces, initialized=True):
    """
    This function takes a list of TLEs and a tensor of times (which must be of same length), and returns the corresponding states.

    Args:
        - tle_sat (``list``): list of TLE objects or TLE object to be propagated
        - tsince (``torch.tensor``): propagation time in minutes (it has to be a tensor of the same size of the list, in case a list of TLEs is passed)
        - initialized (``bool``): whether the TLEs have been initialized or not (default: True)

    Returns:
        - state (``torch.tensor``): (Nx2x3) tensor representing position and velocity in km and km/s, where the first dimension is the batch size.
    """
    from .sgp4_batched import sgp4_batched
    if not initialized:
        initialize_tle(tles)
    state=sgp4_batched(tles, tsinces)
    return state

def propagate(tle, tsinces, initialized=True):
    """
    This function takes a tensor of inputs and a TLE, and returns the corresponding state.
    In particular, multiple behaviors are supported:
    - if a single TLE is provided, then the function returns the state of the satellite at the requested time(s)
    - if a list of TLEs is provided, then the function returns the state of each satellite at the requested times

    In the second case, the length of the list of TLEs must be equal to the length of the tensor of times.

    Args:
        - tle_sat (``list`` or ``dsgp4.tle.TLE``): list of TLE objects or TLE object to be propagated
        - tsince (``torch.tensor``): propagation time in minutes (it has to be a tensor of the same size of the list, in case a list of TLEs is passed)
        - initialized (``bool``): whether the TLEs have been initialized or not (default: True)

    Returns:
        - state (``torch.tensor``): (2x3) tensor representing position and velocity in km and km/s.
    """
    from .sgp4 import sgp4
    if not initialized:
        initialize_tle(tle)
    state=sgp4(tle, tsinces)
    return state

def initialize_tle(tles,
                   gravity_constant_name="wgs-84",
                   with_grad=False):
    """
    This function takes a single `dsgp4.tle.TLE` object or a list of `dsgp4.tle.TLE` objects and initializes the SGP4 propagator.
    This is a necessary step to be ran before propagating TLEs (e.g. before calling `propagate` function).

    Args:
        - tles (``dsgp4.tle.TLE`` or ``list`` of ``dsgp4.tle.TLE``): TLE object or list of TLE objects to be initialized
        - gravity_constant_name (``str``): name of the gravity constant to be used (default: "wgs-84")    
        - with_grad (``bool``): whether to use gradients or not (default: False)
    
    Returns:
        - tle_elements (``torch.tensor``): tensor of TLE parameters (especially useful to retrieve gradients, when `with_grad` is `True`)
    """
    from .sgp4init import sgp4init
    from .sgp4init_batch import sgp4init_batch
    whichconst=get_gravity_constants(gravity_constant_name)
    deep_space_counter=0
    if isinstance(tles,list):
        tle_elements=[]#torch.zeros((len(tles),9),requires_grad=with_grad)
        for tle in tles:
                x=torch.tensor([tle._bstar,
                            tle._ndot,
                            tle._nddot,
                            tle._ecco,
                            tle._argpo,
                            tle._inclo,
                            tle._mo,
                            tle._no_kozai,
                            tle._nodeo
                            ],requires_grad=with_grad)
                tle_elements.append(x)
        xx=torch.stack(tle_elements)
        try:
            tles_batch=tles[0].copy()
            sgp4init_batch(whichconst=whichconst,
                            opsmode='i',
                            satn=tle.satellite_catalog_number,
                            epoch=(tle._jdsatepoch+tle._jdsatepochF)-2433281.5,
                            xbstar=xx[:,0],
                            xndot=xx[:,1],
                            xnddot=xx[:,2],
                            xecco=xx[:,3],
                            xargpo=xx[:,4],
                            xinclo=xx[:,5],
                            xmo=xx[:,6],
                            xno_kozai=xx[:,7],
                            xnodeo=xx[:,8],
                            satellite_batch=tles_batch,
                            )
        except:
            deep_space_counter+=1
        if deep_space_counter>0:
            print("Warning: "+str(deep_space_counter)+" TLEs were not initialized because they are in deep space. Deep space propagation is currently not supported.")
        return tle_elements, tles_batch

    else:
        tle_elements=torch.tensor([tles._bstar,
                                    tles._ndot,
                                    tles._nddot,
                                    tles._ecco,
                                    tles._argpo,
                                    tles._inclo,
                                    tles._mo,
                                    tles._no_kozai,
                                    tles._nodeo
                                    ],requires_grad=with_grad)
        sgp4init(whichconst=whichconst,
                            opsmode='i',
                            satn=tles.satellite_catalog_number,
                            epoch=(tles._jdsatepoch+tles._jdsatepochF)-2433281.5,
                            xbstar=tle_elements[0],
                            xndot=tle_elements[1],
                            xnddot=tle_elements[2],
                            xecco=tle_elements[3],
                            xargpo=tle_elements[4],
                            xinclo=tle_elements[5],
                            xmo=tle_elements[6],
                            xno_kozai=tle_elements[7],
                            xnodeo=tle_elements[8],
                            satellite=tles)
        return tle_elements

def from_year_day_to_date(y,d):
    return (datetime.datetime(y, 1, 1) + datetime.timedelta(d - 1))

def gstime(jdut1):
    tut1 = (jdut1 - 2451545.0) / 36525.0
    temp = -6.2e-6* tut1 * tut1 * tut1 + 0.093104 * tut1 * tut1 + \
         (876600.0*3600 + 8640184.812866) * tut1 + 67310.54841  #  sec
    temp = (temp*(np.pi/180.0) / 240.0) % (2*np.pi) # 360/86400 = 1/240, to deg, to rad

     #  ------------------------ check quadrants ---------------------
    temp=torch.where(temp<0., temp+(2*np.pi), temp)
    return temp

def clone_w_grad(y):
    return y.clone().detach().requires_grad_(True)

def jday(year, mon, day, hr, minute, sec):
    """
    Converts a date and time to a Julian Date. The Julian Date is the number of days since noon on January 1st, 4713 BC.

    Args:
        year (`int`): year
        mon (`int`): month
        day (`int`): day
        hr (`int`): hour
        minute (`int`): minute
        sec (`float`): second

    Returns:
        `float`: Julian Date
    """
    jd=(367.0 * year -
            7.0 * (year + ((mon + 9.0) // 12.0)) * 0.25 // 1.0 +
            275.0 * mon // 9.0 +
            day + 1721013.5)
    fr=(sec + minute * 60.0 + hr * 3600.0) / 86400.0
    return jd,fr

def invjday(jd):
    """
    Converts a Julian Date to a date and time. The Julian Date is the number of days since noon on January 1st, 4713 BC.

    Args:
        jd (`float`): Julian Date

    Returns:
        `tuple`: (year, month, day, hour, minute, second)
    """
    temp    = jd - 2415019.5
    tu      = temp / 365.25
    year    = 1900 + int(tu // 1.0)
    leapyrs = int(((year - 1901) * 0.25) // 1.0)
    days    = temp - ((year - 1900) * 365.0 + leapyrs) + 0.00000000001
    if (days < 1.0):
        year    = year - 1
        leapyrs = int(((year - 1901) * 0.25) // 1.0)
        days    = temp - ((year - 1900) * 365.0 + leapyrs)
    mon, day, hr, minute, sec = days2mdhms(year, days)
    sec = sec - 0.00000086400
    return year, mon, day, hr, minute, sec

def days2mdhms(year, fractional_day):
    """
    Converts a number of days to months, days, hours, minutes, and seconds.

    Args:
        year (`int`): year
        fractional_day (`float`): number of days

    Returns:
        `tuple`: (month, day, hour, minute, second)
    """
    d=datetime.timedelta(days=fractional_day)
    datetime_obj=datetime.datetime(year-1,12,31)+d
    return datetime_obj.month, datetime_obj.day, datetime_obj.hour, datetime_obj.minute, datetime_obj.second+datetime_obj.microsecond/1e6

def from_string_to_datetime(string):
    """
    Converts a string to a datetime object.

    Args:
        string (`str`): string to convert

    Returns:
        `datetime.datetime`: datetime object
    """
    if string.find('.')!=-1:
        return datetime.datetime.strptime(string, '%Y-%m-%d %H:%M:%S.%f')
    else:
        return datetime.datetime.strptime(string, '%Y-%m-%d %H:%M:%S')

def from_mjd_to_epoch_days_after_1_jan(mjd_date):
    """
    Converts a Modified Julian Date to the number of days after 1 Jan 2000.

    Args:
        mjd_date (`float`): Modified Julian Date

    Returns:
        `float`: number of days after 1 Jan 2000
    """
    d = from_mjd_to_datetime(mjd_date)
    dd = d - datetime.datetime(d.year-1, 12, 31)
    days = dd.days
    days_fraction = (dd.seconds + dd.microseconds/1e6) / (60*60*24)
    return days + days_fraction

def from_mjd_to_datetime(mjd_date):
    """
    Converts a Modified Julian Date to a datetime object. The Modified Julian Date is the number of days since midnight on November 17, 1858.

    Args:
        mjd_date (`float`): Modified Julian Date

    Returns:
        `datetime.datetime`: datetime object
    """
    jd_date=mjd_date+2400000.5
    return from_jd_to_datetime(jd_date)

def from_jd_to_datetime(jd_date):
    """
    Converts a Julian Date to a datetime object. The Julian Date is the number of days since noon on January 1st, 4713 BC.

    Args:
        jd_date (`float`): Julian Date

    Returns:
        `datetime.datetime`: datetime object
    """
    year, month, day, hour, minute, seconds=invjday(jd_date)
    e_1=datetime.datetime(year=int(year), month=int(month), day=int(day), hour=int(hour), minute=int(minute), second=0)
    return e_1+datetime.timedelta(seconds=seconds)

def get_non_empty_lines(lines):
    """
    This function returns the non-empty lines of a list of lines.

    Args:
        lines (`list`): list of lines

    Returns:
        `list`: non-empty lines
    """
    if not isinstance(lines, str):
        raise ValueError('Expecting a string')
    lines = lines.splitlines()
    lines = [line for line in lines if line.strip()]
    return lines

def from_datetime_to_fractional_day(datetime_object):
    """
    Converts a datetime object to a fractional day. The fractional day is the number of days since the beginning of the year. For example, January 1st is 0.0, January 2nd is 1.0, etc.

    Args:
        datetime_object (`datetime.datetime`): datetime object to convert

    Returns:
        `float`: fractional day
    """
    d = datetime_object-datetime.datetime(datetime_object.year-1, 12, 31)
    fractional_day = d.days + d.seconds/60./60./24 + d.microseconds/60./60./24./1e6
    return fractional_day

def from_datetime_to_mjd(datetime_obj):
    """
    Converts a datetime object to a Modified Julian Date. The Modified Julian Date is the number of days since midnight on November 17, 1858.

    Args:
        datetime_obj (`datetime.datetime`): datetime object to convert

    Returns:
        `float`: Modified Julian Date
    """
    return from_datetime_to_jd(datetime_obj)-2400000.5

def from_datetime_to_jd(datetime_obj):
    """
    Converts a datetime object to a Julian Date. The Julian Date is the number of days since noon on January 1, 4713 BC.

    Args:
        datetime_obj (`datetime.datetime`): datetime object to convert

    Returns:
        `float`: Julian Date
    """
    return sum(jday(year=datetime_obj.year, mon=datetime_obj.month, day=datetime_obj.day, hr=datetime_obj.hour, minute=datetime_obj.minute, sec=datetime_obj.second+float('0.'+str(datetime_obj.microsecond))))

def from_cartesian_to_tle_elements(state, gravity_constant_name='wgs-72'):
    """
    This function converts the provided state from Cartesian to TLE elements.

    Args:
        state (`np.ndarray`): state to convert
        gravity_constant_name (`str`): name of the central body (default: 'wgs-72')

    Returns:
        tuple: tuple containing: - `float`: semi-major axis - `float`: eccentricity - `float`: inclination - `float`: right ascension of the ascending node - `float`: argument of perigee - `float`: mean anomaly
    """
    _,mu_earth,_,_,_,_,_,_=get_gravity_constants(gravity_constant_name)
    mu_earth=float(mu_earth)*1e9
    kepl_el = from_cartesian_to_keplerian(state, mu_earth)
    tle_elements={}
    tle_elements['mean_motion']         = np.sqrt(mu_earth/((kepl_el[0])**(3.0)))
    tle_elements['eccentricity']        = kepl_el[1]
    tle_elements['inclination']         = kepl_el[2]
    tle_elements['raan']                = kepl_el[3]
    tle_elements['argument_of_perigee'] = kepl_el[4]
    mean_anomaly = kepl_el[5] - kepl_el[1]*np.sin(kepl_el[5])
    tle_elements['mean_anomaly']        = mean_anomaly%(2*np.pi)
    return tle_elements

def from_cartesian_to_keplerian(state, mu):
    """
    This function takes the state in cartesian coordinates and the gravitational
    parameter of the central body, and returns the state in Keplerian elements.

    Args:
        state (`np.array`): numpy array of 2 rows and 3 columns, where
                                    the first row represents position, and the second velocity.
        mu (`float`): gravitational parameter of the central body

    Returns:
        `np.array`: numpy array of the six keplerian elements: (a,e,i,omega,Omega,mean_anomaly)
                                             (i.e., semi major axis, eccentricity, inclination,
                                             right ascension of ascending node, argument of perigee,
                                             mean anomaly). All the angles are in radiants, eccentricity is unitless
                                             and semi major axis is in SI.
    """
    h_bar = np.cross(np.array([state[0,0], state[0,1], state[0,2]]), np.array([state[1,0], state[1,1], state[1,2]]))
    h = np.linalg.norm(h_bar)
    r = np.linalg.norm(np.array([state[0,0], state[0,1], state[0,2]]))
    v = np.linalg.norm(np.array([state[1,0], state[1,1], state[1,2]]))
    E = 0.5*(v**2)-mu/r
    a = -mu/(2*E)
    e = np.sqrt(1-(h**2)/(a*mu))
    i = np.arccos(h_bar[2]/h)
    Omega = np.arctan2(h_bar[0],-h_bar[1])

    lat = np.arctan2(np.divide(state[0,2],(np.sin(i))), (state[0,0]*np.cos(Omega) + state[0,1]*np.sin(Omega)))
    p = a*(1-e**2)
    nu = np.arctan2(np.sqrt(p/mu)*np.dot(np.array([state[0,0], state[0,1], state[0,2]]),np.array([state[1,0], state[1,1], state[1,2]])), p-r)
    omega = (lat-nu)
    eccentric_anomaly = 2*np.arctan(np.sqrt((1-e)/(1+e))*np.tan(nu/2))
    n = np.sqrt(mu/(a**3))
    mean_anomaly=eccentric_anomaly-e*np.sin(eccentric_anomaly)
    #I make sure they are always in 0,2pi
    if mean_anomaly<0:
        mean_anomaly = 2*np.pi-abs(mean_anomaly)
    if omega<0:
        omega=2*np.pi-abs(omega)
    if Omega<0:
        Omega=2*np.pi-abs(Omega)
    if abs(mean_anomaly)>2*np.pi:
        mean_anomaly=mean_anomaly%(2*np.pi)
    if abs(omega)>2*np.pi:
        omega=omega%(2*np.pi)
    if abs(Omega)>2*np.pi:
        Omega=Omega%(2*np.pi)
    return np.array([a, e, i, Omega, omega, mean_anomaly])

def from_cartesian_to_keplerian_torch(state, mu):
    """
    Same as from_cartesian_to_keplerian, but for torch tensors.

    Args:
        state (`np.array`): numpy array of 2 rows and 3 columns, where
                                    the first row represents position, and the second velocity.
        mu (`float`): gravitational parameter of the central body

    Returns:
        `np.array`: numpy array of the six keplerian elements: (a,e,i,omega,Omega,mean_anomaly)
                                             (i.e., semi major axis, eccentricity, inclination,
                                             right ascension of ascending node, argument of perigee,
                                             mean anomaly). All the angles are in radiants, eccentricity is unitless
                                             and semi major axis is in SI.
    """
    h_bar = torch.cross(state[0], state[1])
    h = h_bar.norm()
    r = state[0].norm()
    v = torch.norm(state[1])
    E = 0.5*(v**2)-mu/r
    a = -mu/(2*E)
    e = torch.sqrt(1-(h**2)/(a*mu))
    i = torch.arccos(h_bar[2]/h)
    Omega = torch.arctan2(h_bar[0],-h_bar[1])
    lat = torch.arctan2(torch.divide(state[0,2],(torch.sin(i))), (state[0,0]*torch.cos(Omega) + state[0,1]*torch.sin(Omega)))
    p = a*(1-e**2)
    nu = torch.arctan2(torch.sqrt(p/mu)*torch.dot(state[0],state[1]), p-r)
    omega = (lat-nu)
    eccentric_anomaly = 2*torch.arctan(torch.sqrt((1-e)/(1+e))*torch.tan(nu/2))
    n = torch.sqrt(mu/(a**3))
    mean_anomaly=eccentric_anomaly-e*torch.sin(eccentric_anomaly)
    #I make sure they are always in 0,2pi
    mean_motion=torch.sqrt(mu/((a)**(3.0)))
    xpdotp   =  1440.0 / (2.0 *np.pi)
    no_kozai_conversion_factor=xpdotp/43200.0* np.pi
    no_kozai=mean_motion/no_kozai_conversion_factor
    return [no_kozai, e, i, Omega, omega, mean_anomaly]
